# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Timba.TDL import *
from Meow.Direction import radec_to_lmn
import Meow
from Meow import Context
from Meow import StdTrees
from Meow import ParmGroup

import math
from math import sqrt,atan2

DEG = math.pi/180.;
ARCMIN = DEG/60;
ARCSEC = DEG/3600;

class circular_aperture_beam (object):
  label = "circular_aperture";
  name  = "Circular aperture beam model";
  
  def __init__ (self):
    self._options = [
      TDLOption('bf',"Beam factor (1/GHz)",[65.],more=float,namespace=self),
      TDLOption('dish_sizes',"Dish size(s)",[25],more=str,namespace=self)
    ];
    self._prepared = False;
    
  def option_list (self):
    return self._options;
  
  def prepare (self):
    if self._prepared:
      return;
    try:
      self._sizes = map(float,str(self.dish_sizes).split());
    except:
      raise RuntimeError,"illegal dish sizes specification";
    self._per_station = len(self._sizes)>1;
    self._prepared = True;

  def compute (self,E,lm,p=0):
    """Circular apertures -> Airy disk"""
    ns = E.Subscope();
    if isinstance(lm,(list,tuple)):
      l,m = lm;
      r = ns.r << sqrt(l*l+m*m);
    else:
      r = ns.r << Meq.Norm(lm);
    size = self._sizes[min(p,len(self._sizes)-1)];
    ns.theta_r << r * size * math.pi * self.bf * 1e-9 * Meq.Freq() + 1e-6;
    ns.E_theta << Meq.Divide(2 * Meq.Bessel(ns.theta_r, order=1), ns.theta_r);
    E << Meq.Matrix22(ns.E_theta, 0, 0, ns.E_theta);
    return E;

def compute_jones (Jones,sources,stations=None,
                    label="beam",pointing_offsets=None,inspectors=[],
                    solvable_sources=set(),**kw):
  """Computes beam gain for a list of sources.
  The output node, will be qualified with either a source only, or a source/station pair
  """;
  stations = stations or Context.array.stations();
  ns = Jones.Subscope();
  
  # figure out beam model
  for beam_model in MODELS:
    if globals().get('use_%s'%model.label,None):
      break;
  else:
    raise RuntimeError,"no beam model selected";
  beam_model.prepare();
  
  # this dict will hold LM tuples (or nodes) for each source.
  lmsrc = {};
  # see if sources have a "beam_lm" attribute, use that for beam offsets
  for src in sources:
    lm = src.get_attr("beam_lm",None) or src.get_attr("_lm_ncp",None);
    if lm:
      l,m = lmsrc[src.name] = lm;
      src.set_attr(label+'r',math.sqrt(l**2+m**2)/math.pi*(180*60));
    # else try to use static lm coordinates
    else:
      # else try to use static lm coordinates
      lmnst = src.direction.lmn_static();
      if lmnst:
        l,m = lmsrc[src.name] = lmnst[0:2];
        src.set_attr(label+'r',math.sqrt(l**2+m**2)/math.pi*(180*60));
      # else use lmn node
      else:
        lmsrc[src.name] = src.direction.lm();
    
  # set of all sources, will later become set of sources for which
  # a beam hasn't been computed
  all_sources = set([src.name for src in sources]);
  
  # if pointing errors are enabled, compute beams for sources with a PE
  if pointing_offsets:
    # is a subset specified?
    if pe_subset != "all":
      pe_sources = set(pe_subset.split(" "));
      # make sure pe_sources does not contain unknown sources
      pe_sources &= all_sources;
      # remove PE_enabled sources from global list
      all_sources -= pe_sources;
    # else all sources have a PE
    else:
      pe_sources = all_sources;
      all_sources = [];
    # create nodes to compute actual pointing per source, per antenna
    for name in pe_sources:
      solvable_sources.add(name);
      lm = lmsrc[name];
      # if LM is a static constant still, make constant node for it
      if isinstance(lm,(list,tuple)):
        lm = ns.lm(name) << Meq.Constant(value=Timba.array.array(lm));
      for ip,p in enumerate(stations):
        # make offset lm
        lm1 = lm(p) << lm + pointing_offsets(p);
        # make beam model
        beam_model.compute(Jones(name,p),lm1,ip);
  
  # now, all_sources is the set of sources for which we haven't computed the beam yet
  if all_sources:
    for name in all_sources:
      for ip,p in enumerate(stations):
        beam_model.compute(Jones(name,p),lmsrc[name],ip);
          
  return Jones;

# this will be set to True on a per-source basis, or if beam scale is solvable
solvable = False;

# available beam models
MODELS = [
  cicular_aperture_beam()
];


_model_option = TDLCompileMenu("Beam model",
  exclusive='model_type',
  *[ TDLMenu(model.name,toggle="use_%s"%model.label,*model.option_list())
      for model in MODELS ]);

TDLCompileOption('pe_subset',"Apply pointing errors (if any) to a subset of sources",
        ["all"],more=str,doc="""Selects a subset of sources to which pointing errors 
        (if any) are applied. Enter soure names separated by space""");


