#!/usr/bin/env python

import numpy as np
from matplotlib import pyplot as plt
import sys
from sched_sens import antenna_sefd_dict
from scipy.special import jn

dish_sizes = {'Ro70': 70, 'Nl': 25, 'Jb2': 25, 'Jb1': 76, 'La': 25, 'Tr': 32, 'Pb': -1, 'Ny': 20, 'Sv': 32, 'Ro34': 34, 'Pv': -1, 'Kp': 25, 'Nt': 32, 'Tc': -1, 'Pt': 25, 'Y27': 3000, 'Bd': 32, 'Wb': 2700, 'Hh': 26, 'Hn': 25, 'Br': 25, 'Fd': 25, 'Ys': 40, 'Wz': 20, 'On': 25, 'Ka': -1, 'Mc': 32, 'Mh': 14, 'Km': 40, 'Ur': 25, 'Go': -1, 'Sh': 25, 'Mk': 25, 'Sm': -1, 'Sc': 25, 'Ov': 25, 'My': 50, 'Cm': 32, 'Ef': 100, 'Zc': 32, 'Ar': 305, 'Gb': 100, 'W1': 25, 'Y1': 25}

POINTS = 1000	# Number of samples
RADIUS = 1.0	# Degrees, field of view to sample
imsize = 512	# Size of 2D plot
wl = 0.18	# Metres, wavelength

# Yurii's NME without Wb:
nme_ant_list = ['Ef', 'Jb1', 'On', 'Mc', 'Tr', 'Sv', 'Zc', 'Bd', 'Ur', 'Sh', 'Hh']
# VLBA
vlba_ant_list = ['Br', 'Fd','Hn','Kp','La','Mk','Nl','Ov','Pt','Sc']

def jinc(x):
	return 2 * jn(1, np.pi * x) / (np.pi * x)

def comb_sefd(sefd_dict, ant_list):
	comb = 0
	for i in range(len(ant_list)-1):
		for j in range(i+1 ,len(ant_list)):
			comb += 1.0 / (sefd_dict[ant_list[i]] * sefd_dict[ant_list[j]])
	sefd_total = 1.0 / comb ** 0.5
	return sefd_total

def baseline_sefd(sefd_dict, ant1, ant2):
	return (sefd_dict[ant1] * sefd_dict[ant2]) ** 0.5

def mirror(a):
	return np.concatenate((a[::-1], a))

def neg_mirror(a):
	return np.concatenate((-a[::-1], a))

def ant_beam(rad_array, dish_size, wavelength):
	return abs(jinc(rad_array * dish_size / wavelength))

def gen_beams(ant_list, sefd_dict, rad_array):
	beams = []
	comb_beam = np.zeros(shape=rad_array.shape)
	for i in range(len(ant_list)-1):
		for j in range(i+1, len(ant_list)):
			ant1_beam = ant_beam(rad_array, dish_sizes[ant_list[i]], wl)
			ant2_beam = ant_beam(rad_array, dish_sizes[ant_list[j]], wl)
			bl_sefd = baseline_sefd(sefd_dict, ant_list[i], ant_list[j])
			g = ant1_beam * ant2_beam / bl_sefd
			beams.append(g)
			comb_beam += g ** 2
	comb_beam **= 0.5
	beams.append(comb_beam)
	return beams

def plot_beam2d(beam, imsize):
	max_a = 2 ** 0.5 * imsize
	beam2d = np.empty(shape=(imsize, imsize), dtype=np.float32)
	halfim = imsize / 2
	for i in range(halfim):
		for j in range(halfim):
			a = (i**2 + j**2) ** 0.5
			val = beam[a/max_a * POINTS]
			beam2d[halfim+i, halfim+j] = val
			beam2d[halfim-i, halfim+j] = val
			beam2d[halfim+i, halfim-j] = val
			beam2d[halfim-i, halfim-j] = val
	return beam2d

r = np.radians(RADIUS)
ang = np.arange(0, r, r / POINTS)
sefd_dict = antenna_sefd_dict(wl)
print 'NME SEFD:', comb_sefd(sefd_dict, nme_ant_list)
print 'VLBA SEFD:', comb_sefd(sefd_dict, vlba_ant_list)
		
plt.subplot(231)
beam_list = gen_beams(nme_ant_list, sefd_dict, ang)
nme_beam = beam_list[-1]
for beam in beam_list:
	plt.plot(neg_mirror(60 * np.degrees(ang)), mirror(beam))

plt.subplot(232)
sq_max = RADIUS * 60 / 2 ** 0.5
plt.imshow(plot_beam2d(beam_list[-1], imsize), extent=(-sq_max, sq_max, -sq_max, sq_max))

plt.subplot(234)
beam_list = gen_beams(vlba_ant_list, sefd_dict, ang)
vlba_beam = beam_list[-1]
for beam in beam_list:
	plt.plot(neg_mirror(60 * np.degrees(ang)), mirror(beam))

plt.subplot(235)
sq_max = RADIUS * 60 / 2 ** 0.5
plt.imshow(plot_beam2d(beam_list[-1], imsize), extent=(-sq_max, sq_max, -sq_max, sq_max))

plt.subplot(233)
plt.plot(neg_mirror(60 * np.degrees(ang)), mirror(nme_beam))
plt.plot(neg_mirror(60 * np.degrees(ang)), mirror(vlba_beam))

plt.show()
