#!/usr/bin/env ParselTongue
#
# Tool to convert AIPS SN tables into Measurement Set Calibration tables
# Stephen Bourke, JIVE

import AIPS
from AIPSData import AIPSCat
from AIPSData import AIPSUVData
import Wizardry.AIPSData as wiz
from pyrap.tables import table
from pyrap.quanta import quantity
import math
import sys
import shutil
import datetime
import calendar
import numpy
import os.path

mod_jul_day = quantity(40587.0, 'd')
mod_jul_day.convert(quantity(1.0, 's')) # Modified Julien day

if len(sys.argv) != 7 and len(sys.argv) != 8:
	print >> sys.stderr, 'Usage: %s <userno> <disk> <cno> <bpver> <ms> <outcal> [antfile]' % sys.argv[0]
	sys.exit()
AIPS.userno, disk, cno, bpver = [int(i) for i in sys.argv[1:5]]
ms_name, cal_base = sys.argv[5:7]
try:
	antfile = sys.argv[7]
except IndexError:
	antfile = None

ds = [i for i in AIPSCat()[disk] if i['cno'] == cno][0]
uv = wiz.AIPSUVData(ds['name'], ds['klass'], disk, ds['seq'])
obs_date = datetime.datetime(*[int(i) for i in uv.header.date_obs.split('-')])

print AIPSUVData(uv), 'BP%d' % bpver, '->', ms_name, cal_base

bp = uv.table('BP', bpver)
pols = bp.keywords['NO_POL']
ifs = bp.keywords['NO_IF']	# AIPS IF / Casa SPW
chans = bp.keywords['NO_CHAN']

ms_ant = table(ms_name + '/ANTENNA', ack=False)
ms_num_ants = len(ms_ant)
ms_ant.close()

class progress:
	def __init__(self, done_value=100.0, dot_value=2.51):
		self.done_value = done_value
		self.progress_next = 0
		self.dot_value = dot_value
		self.progress_dot = dot_value

	def update(self, value):
		progress_curr = float(value) / self.done_value * 100
		if progress_curr > self.progress_next:
			sys.stdout.write('%d%%' % self.progress_next)
			sys.stdout.flush()
			self.porgress_dot = self.progress_next + self.dot_value
			self.progress_next += 10
		if progress_curr > self.progress_dot:
			sys.stdout.write('.')
			sys.stdout.flush()
			self.progress_dot += self.dot_value

	def done(self):
		sys.stdout.write('100%\n')
		sys.stdout.flush()

# User specified mapping from AIPS ants to MS ant ids
ant_aips2ms = {}
if antfile:
	inf = open(antfile)
	for line in inf:
		aips_ant_no, ms_ant_id = [int(x) for x in line.split()]
		ant_aips2ms[aips_ant_no] = ms_ant_id
	inf.close()
else:
	for i in range(1,100):
		ant_aips2ms[i] = i - 1

prgs = progress(len(bp))
sys.stdout.write('Reading BP table\n')
sys.stdout.flush()
cals = {}
for i in range(len(bp)):
	prgs.update(i)
	sol_time = obs_date + datetime.timedelta(float(bp[i]['time']))
	time_unix = calendar.timegm(sol_time.timetuple())
	time_casa = time_unix + mod_jul_day.get_value()
	if time_casa not in cals:
		cals[time_casa] = {}
	time_interval = quantity(bp[i]['interval'], 'd')
	time_interval.convert(quantity(1.0, 's'))
	time_interval = time_interval.get_value()
	field_id = bp[i]['source_id'] - 1
	try:
		antenna1 = ant_aips2ms[bp[i]['antenna']]
	except KeyError:
		continue
	if antenna1 not in cals[time_casa]:
		cals[time_casa][antenna1] = {}
	for iif in range(ifs):
		if not iif in cals[time_casa][antenna1]:
			cals[time_casa][antenna1][iif] = {}
		row = iif * len(bp) + i
		gain = numpy.ones(shape=(chans, pols), dtype='complex64')
		snr = numpy.ones(shape=(chans,pols), dtype='float32')
		for p in range(pols):
			gain[:,p].real = bp[i]['real_%d' % (p+1)][iif*chans:(iif+1)*chans]
			gain[:,p].imag = bp[i]['imag_%d' % (p+1)][iif*chans:(iif+1)*chans]
			if ifs == 1:
				snr[:,p] = bp[i]['weight_%d' % (p+1)]
			else:
				snr[:,p] = bp[i]['weight_%d' % (p+1)][iif]
		cals[time_casa][antenna1][iif]['common'] = {
			   'TIME': time_casa, 'INTERVAL': time_interval,
			   'SNR': snr, 'ANTENNA1': antenna1,
			   'CAL_DESC_ID': iif, 'FIELD_ID': field_id}
		cals[time_casa][antenna1][iif]['gain'] = {
			   'GAIN': gain}
prgs.done()

for cal_type in ['bcal']:
	cal_table = cal_base + '.' + cal_type
	shutil.copytree('empty.' + cal_type, cal_table)
	# Make an CAL_DESC entry for each AIPS IF
	cal_desc_defs = {'NUM_SPW': 1,
			 'NUM_CHAN': numpy.ones(shape=(chans,), dtype='int32'),
			 'NUM_RECEPTORS': pols,
			 'N_JONES': 2,
			 'CHAN_FREQ': numpy.zeros(shape=(chans,1), dtype='float64'),
			 'MEAS_FREQ_REF': 0,
			 'CHAN_WIDTH': numpy.zeros(shape=(chans,1), dtype='float64'),
			 'CHAN_RANGE': numpy.zeros(shape=(chans,1,2), dtype='int32'),
			 'POLARIZATION_TYPE': {'shape': [chans, 1], 'array': chans * ['']},
			 'JONES_TYPE': 'full',
			 'MS_NAME': os.path.basename(ms_name)}
	
	cal_desc = table(cal_table + '/CAL_DESC', ack=False, readonly=False)
	cal_desc.addrows(ifs)
	for i in range(ifs):
		cal_desc[i] = cal_desc_defs
		cal_desc[i] = {'SPECTRAL_WINDOW_ID': numpy.array((i,), dtype='int32')}
		cal_desc[i].update()
	cal_desc.close()

# Make the CAL_MAIN table
defs = {'TOTAL_SOLUTION_OK': True,
        'TOTAL_FIT_WEIGHT': 1.0,
	'TOTAL_FIT': 1.0,
	'FIT': numpy.ones(shape=(chans,1), dtype='float32'),
	'FIT_WEIGHT': numpy.ones(shape=(chans,1), dtype='float32')}

defs_good = {'SOLUTION_OK': numpy.ones(shape=(chans,pols), dtype='bool'),
             'FLAG': numpy.zeros(shape=(chans,pols), dtype='bool')}

defs_bad = {'SOLUTION_OK': numpy.zeros(shape=(chans,pols), dtype='bool'),
            'FLAG': numpy.ones(shape=(chans,pols), dtype='bool'),
	    'GAIN': numpy.ones(shape=(chans, pols), dtype='complex64'),
	    'SNR': numpy.zeros(shape=(chans,pols), dtype='float32')}

bcal = table(cal_base + '.bcal', ack=False, readonly=False)
bcal.addrows(ms_num_ants * len(cals) * ifs)

timestamps = cals.keys()
timestamps.sort()
i = 0
prgs = progress(len(bcal))
print 'Writing cal tables'
for iif in range(ifs):
	for ts in timestamps:
		for ant in range(ms_num_ants):
			try:
				c_common = cals[ts][ant][iif]['common']
				c_gain = cals[ts][ant][iif]['gain']
				defs2 = defs_good
			except KeyError:
				# Use another solution as a base
				dummy_ant = cals[ts].keys()[0]
				dummy_if = cals[ts][dummy_ant].keys()[0]
				c_common = cals[ts][dummy_ant][dummy_if]['common']
				c_gain = cals[ts][dummy_ant][dummy_if]['gain']
				c_common['ANTENNA1'] = ant
				c_common['CAL_DESC_ID'] = iif
				defs2 = defs_bad
			bcal[i] = c_common
			bcal[i] = c_gain
			bcal[i] = defs
			bcal[i] = defs2
			bcal[i].update()
			i += 1
			prgs.update(i)
prgs.done()
