#!/usr/bin/env ParselTongue
#
# Tool to convert AIPS SN tables into Measurement Set Calibration tables
# Stephen Bourke, JIVE

import AIPS
from AIPSData import AIPSCat
from AIPSData import AIPSUVData
import Wizardry.AIPSData as wiz
from pyrap.tables import table
from pyrap.quanta import quantity
import math
import sys
import shutil
import datetime
import calendar
import numpy
import os.path

mod_jul_day = quantity(40587.0, 'd')
mod_jul_day.convert(quantity(1.0, 's')) # Modified Julien day

if len(sys.argv) != 7 and len(sys.argv) != 8:
	print >> sys.stderr, 'Usage: %s <userno> <disk> <cno> <clver> <ms> <outcal> [antfile]' % sys.argv[0]
	sys.exit()
AIPS.userno, disk, cno, clver = [int(i) for i in sys.argv[1:5]]
ms_name, cal_base = sys.argv[5:7]
try:
	antfile = sys.argv[7]
except IndexError:
	antfile = None

ds = [i for i in AIPSCat()[disk] if i['cno'] == cno][0]
uv = wiz.AIPSUVData(ds['name'], ds['klass'], disk, ds['seq'])
obs_date = datetime.datetime(*[int(i) for i in uv.header.date_obs.split('-')])

print AIPSUVData(uv), 'CL%d' % clver, '->', ms_name, cal_base

cl = uv.table('CL', clver)
pols = cl.keywords['NO_POL']
ifs = cl.keywords['NO_IF']	# AIPS IF / Casa SPW

ms_ant = table(ms_name + '/ANTENNA', ack=False)
ms_num_ants = len(ms_ant)
ms_ant.close()

class progress:
	def __init__(self, done_value=100.0, dot_value=2.51):
		self.done_value = done_value
		self.progress_next = 0
		self.dot_value = dot_value
		self.progress_dot = dot_value

	def update(self, value):
		progress_curr = float(value) / self.done_value * 100
		if progress_curr > self.progress_next:
			sys.stdout.write('%d%%' % self.progress_next)
			sys.stdout.flush()
			self.porgress_dot = self.progress_next + self.dot_value
			self.progress_next += 10
		if progress_curr > self.progress_dot:
			sys.stdout.write('.')
			sys.stdout.flush()
			self.progress_dot += self.dot_value

	def done(self):
		sys.stdout.write('100%\n')
		sys.stdout.flush()

# User specified mapping from AIPS ants to MS ant ids
ant_aips2ms = {}
if antfile:
	inf = open(antfile)
	for line in inf:
		aips_ant_no, ms_ant_id = [int(x) for x in line.split()]
		ant_aips2ms[aips_ant_no] = ms_ant_id
	inf.close()

prgs = progress(len(cl))
sys.stdout.write('Reading CL table\n')
sys.stdout.flush()
cals = {}
for i in range(len(cl)):
	prgs.update(i)
	sol_time = obs_date + datetime.timedelta(float(cl[i]['time']))
	time_unix = calendar.timegm(sol_time.timetuple())
	time_casa = time_unix + mod_jul_day.get_value()
	if time_casa not in cals:
		cals[time_casa] = {}
	time_interval = quantity(cl[i]['time_interval'], 'd')
	time_interval.convert(quantity(1.0, 's'))
	time_interval = time_interval.get_value()
	field_id = cl[i]['source_id'] - 1
	if antfile:
		antenna1 = ant_aips2ms[cl[i]['antenna_no']]
	else:
		antenna1 = cl[i]['antenna_no'] - 1
	if antenna1 not in cals[time_casa]:
		cals[time_casa][antenna1] = {}
	for iif in range(ifs):
		if not iif in cals[time_casa][antenna1]:
			cals[time_casa][antenna1][iif] = {}
		row = iif * len(cl) + i
		gain = numpy.ones(shape=(1, pols), dtype='complex64')
		mbdelay = numpy.ones(shape=(1, pols), dtype='complex64')
		sbdelay = numpy.ones(shape=(1, pols), dtype='complex64')
		snr = numpy.ones(shape=(1,pols), dtype='float32')
		for p in range(pols):
			if ifs == 1:
				gain[0][p] = cl[i]['real%d' % (p+1)] + 1j * cl[i]['imag%d' % (p+1)]
				sbdelay[0][p] = cl[i]['delay_%d' % (p+1)] * 1e9
				snr[0][p] = cl[i]['weight_%d' % (p+1)]
			else:
				gain[0][p] = cl[i]['real%d' % (p+1)][iif] + 1j * cl[i]['imag%d' % (p+1)][iif]
				sbdelay[0][p] = cl[i]['delay_%d' % (p+1)][iif] * 1e9
				snr[0][p] = cl[i]['weight_%d' % (p+1)][iif]
			mbdelay[0][p] = cl[i]['mbdelay%d' % (p+1)] * 1e9
		cals[time_casa][antenna1][iif]['common'] = {
			   'TIME': time_casa, 'INTERVAL': time_interval,
			   'SNR': snr, 'ANTENNA1': antenna1,
			   'CAL_DESC_ID': iif, 'FIELD_ID': field_id}
		cals[time_casa][antenna1][iif]['gcal'] = {
			   'GAIN': gain}
		cals[time_casa][antenna1][iif]['mbdcal'] = {
			   'GAIN': mbdelay}
		cals[time_casa][antenna1][iif]['sbdcal'] = {
			   'GAIN': sbdelay}
prgs.done()

for cal_type in ['gcal', 'mbdcal', 'sbdcal']:
	cal_table = cal_base + '.' + cal_type
	shutil.copytree('empty.' + cal_type, cal_table)
	# Make an CAL_DESC entry for each AIPS IF
	cal_desc_defs = {'NUM_SPW': 1,
			 'NUM_CHAN': numpy.ones(shape=(1,), dtype='int32'),
			 'NUM_RECEPTORS': pols,
			 'N_JONES': 2,
			 'CHAN_FREQ': numpy.zeros(shape=(1,1), dtype='float64'),
			 'MEAS_FREQ_REF': 0,
			 'CHAN_WIDTH': numpy.zeros(shape=(1,1), dtype='float64'),
			 'CHAN_RANGE': numpy.zeros(shape=(1,1,2), dtype='int32'),
			 'POLARIZATION_TYPE': {'shape': [1, 1], 'array': ['']},
			 'JONES_TYPE': 'full',
			 'MS_NAME': os.path.basename(ms_name)}
	
	cal_desc = table(cal_table + '/CAL_DESC', ack=False, readonly=False)
	cal_desc.addrows(ifs)
	for i in range(ifs):
		cal_desc[i] = cal_desc_defs
		cal_desc[i] = {'SPECTRAL_WINDOW_ID': numpy.array((i,), dtype='int32')}
		cal_desc[i].update()
	cal_desc.close()

# Make the CAL_MAIN table
defs = {'TOTAL_SOLUTION_OK': True,
        'TOTAL_FIT_WEIGHT': 1.0,
	'TOTAL_FIT': 1.0,
	'FIT': numpy.ones(shape=(1,1), dtype='float32'),
	'FIT_WEIGHT': numpy.ones(shape=(1,1), dtype='float32')}

defs_good = {'SOLUTION_OK': numpy.ones(shape=(1,pols), dtype='bool'),
             'FLAG': numpy.zeros(shape=(1,pols), dtype='bool')}

defs_bad = {'SOLUTION_OK': numpy.zeros(shape=(1,pols), dtype='bool'),
            'FLAG': numpy.ones(shape=(1,pols), dtype='bool'),
	    'GAIN': numpy.ones(shape=(1, pols), dtype='complex64'),
	    'SNR': numpy.zeros(shape=(1,pols), dtype='float32')}

gcal = table(cal_base + '.gcal', ack=False, readonly=False)
gcal.addrows(ms_num_ants * len(cals) * ifs)
mbdcal = table(cal_base + '.mbdcal', ack=False, readonly=False)
mbdcal.addrows(ms_num_ants * len(cals) * ifs)
sbdcal = table(cal_base + '.sbdcal', ack=False, readonly=False)
sbdcal.addrows(ms_num_ants * len(cals) * ifs)

timestamps = cals.keys()
timestamps.sort()
i = 0
prgs = progress(len(gcal))
print 'Writing cal tables'
for iif in range(ifs):
	for ts in timestamps:
		for ant in range(ms_num_ants):
			try:
				c_common = cals[ts][ant][iif]['common']
				c_gain = cals[ts][ant][iif]['gcal']
				c_mbdelay = cals[ts][ant][iif]['mbdcal']
				c_sbdelay = cals[ts][ant][iif]['sbdcal']
				defs2 = defs_good
			except KeyError:
				# Use another solution as a base
				dummy_ant = cals[ts].keys()[0]
				dummy_if = cals[ts][dummy_ant].keys()[0]
				c_common = cals[ts][dummy_ant][dummy_if]['common']
				c_gain = cals[ts][dummy_ant][dummy_if]['gcal']
				c_mbdelay = cals[ts][dummy_ant][dummy_if]['mbdcal']
				c_sbdelay = cals[ts][dummy_ant][dummy_if]['sbdcal']
				c_common['ANTENNA1'] = ant
				c_common['CAL_DESC_ID'] = iif
				defs2 = defs_bad
			gcal[i] = c_common
			gcal[i] = c_gain
			gcal[i] = defs
			gcal[i] = defs2
			gcal[i].update()
			mbdcal[i] = c_common
			mbdcal[i] = c_mbdelay
			mbdcal[i] = defs
			mbdcal[i] = defs2
			mbdcal[i].update()
			sbdcal[i] = c_common
			sbdcal[i] = c_sbdelay
			sbdcal[i] = defs
			sbdcal[i] = defs2
			sbdcal[i].update()
			i += 1
			prgs.update(i)
prgs.done()
