#!/usr/bin/env python
#
# Program to geometrically shift a field in a Measurement Set
# Stephen Bourke, jive
# Dec, 2010

import math
import numpy
import cmath
import os.path
import progress
from taskinit import casalog, qa
from casa_utils import table, get_field_id, subtable

stokes = [None,
	'I', 'Q', 'U', 'V',
	'RR', 'RL', 'LR', 'LL',
	'XX', 'XY', 'YX', 'YY']

# I = XX + YY
# Q = XX - YY
# U = XY + YX
# V = (XY - YX)/1j

circs = ['RR', 'RL', 'LR', 'LL']
lins = ['XX', 'XY', 'YX', 'YY']

def pol_type(arr):
	is_circ = False not in [i in circs for i in arr]
	is_lin = False not in [i in lins for i in arr]

	if is_circ:
		return 'circ'
	elif is_lin:
		return 'linear'
	else:
		return 'unknown'

def pol_code(x):
	# pol strings -> pol codes
	if isinstance(x, str):
		return stokes.index(x)
	else:
		return [stokes.index(i) for i in x]

def pol_str(x):
	# pol codes -> pol strings
	if hasattr(x, '__getitem__'):
		return [stokes[i] for i in x]
	else:
		return stokes[x]

lin_codes = pol_codes(['XX', 'XY', 'YX', 'YY'])

def lin2stokes(msname, field, shiftval, nrows=2**10):
	"""Geometrically shift a field in a Measurement Set.

	'field' can be field ID or name.
	'shiftval' is a RA, DEC measure, eg: ['1arcsec', '1arcsec']
	'nrows' is the number of rows processed per loop,
	 must fit in memory, defaults to 1024."""

	casalog.origin('chstokes')
	ms = table(msname, readonly=False)
	pol = subtable(ms, 'POLARIZATION', readonly=False)
	if pol.nrows() != 1:
		casalog.post('Multiple polarization setups present. Bailing out')
		return
	pols = pol.getcell('CORR_TYPE', 0)
	assert list(pols) == [9, 10, 11, 12]	#FIXME
	
	# Rotate UVWs and update phase
	startrow = 0
	vis_data = subt.getcol('DATA', startrow, nrows)
	vis_new = numpy.empty(shape=vis_data.shape, dtype=vis_data.dtype)
	casalog.post('Converting stokes')
	shift_progress = progress.progress(subt.nrows())
	for i in range(subt.nrows()):
		if i >= startrow + nrows:
			# Write back data and get next set of rows
			subt.putcol('DATA', vis_new, startrow, nrows)
			startrow += nrows
			vis_data = subt.getcol('DATA', startrow, nrows)
		# FIXME
		vis_new[0,:,:] = vis_data[0,:,:] + vis_data[3,:,:]
		vis_new[1,:,:] = vis_data[0,:,:] - vis_data[3,:,:]
		vis_new[2,:,:] = vis_data[1,:,:] + vis_data[2,:,:]
		vis_new[3,:,:] = (vis_data[1,:,:] - vis_data[2,:,:]) / 1j
		shift_progress.update(i)
	subt.putcol('DATA', vis_new, startrow, nrows)
	shift_progress.done()
	ms.close()
	
	# Update position in FIELD subtable
	casalog.post('Updating data in POLARIZATION table.')
	pols[:] = [1, 2, 3, 4] # FIXME
	pol.putcell('COLL_TYPE', 0, pols)
	pol.close()
