#!/usr/bin/env ParselTongue
#
# Tool to convert AIPS Calibration tables to Measurement Set Calibration tables
# Stephen Bourke, JIVE
# Version 2010.11.29

import AIPS
from AIPSData import AIPSCat
from AIPSData import AIPSUVData
import Wizardry.AIPSData as wiz
from casa_utils import table
from cal_table import create_cal
from progress import progress
import sys
import shutil
import datetime
import calendar
import numpy
import os.path
from taskinit import casalog
import empty_cal

mod_jul_day = 3506716800.0

def importaipscal(vis, caltable, userno, indisk, cno, inver, mode, antfile):
	if len(AIPS.disks) == 1:
		print >> sys.stderr, 'No AIPS disks. Use AIPSLite.attach_disk'
		return
	if mode not in ['CL', 'SN', 'BP']:
		print >> sys.stderr, "Invalid mode: '%s'. Must be 'CL', 'SN' or 'BP'." % mode
		return
	cal_mode = mode
	DOBAND = cal_mode == 'BP'
	DOCALIB = not DOBAND

	AIPS.userno = userno
	ms_name = vis
	cal_base = caltable

	# Select dataset and table
	ds = [i for i in AIPSCat()[indisk] if i['cno'] == cno][0]
	uv = wiz.AIPSUVData(ds['name'], ds['klass'], indisk, ds['seq'])
	obs_date = datetime.datetime(*[int(i) for i in uv.header.date_obs.split('-')])
	print AIPSUVData(uv), '%s%d' % (cal_mode, inver), '->', ms_name, cal_base
	
	cal_aips = uv.table(cal_mode, inver)
	pols = cal_aips.keywords['NO_POL']
	ifs = cal_aips.keywords['NO_IF']	# AIPS IF / Casa SPW
	if DOBAND:
		chans = cal_aips.keywords['NO_CHAN']
	else:
		chans = 1
	
	# User specified mapping from AIPS ants to MS ant ids
	ant_mapping = {}
	if antfile:
		inf = open(antfile)
		for line in inf:
			aips_ant_no, ms_ant_id = [int(x) for x in line.split()]
			ant_mapping[aips_ant_no] = ms_ant_id
		inf.close()
	
	def ant_aips2ms(ant_num):
		if ant_mapping:
			return ant_mapping[ant_num]
		else:
			return ant_num - 1
	
	prgs = progress(len(cal_aips))
	sys.stdout.write('Reading %s table\n' % cal_mode)
	sys.stdout.flush()
	
	# Holds all table data as a multi-level dict indexed by time, antenna,
	# IF, mode. Where mode is one of 'common', 'sbdelay', 'mbdelay'. Format
	# is the same as in the final Table. Casa requires a fairly regular
	# table with rows present for all antennas, and IFs. 
	cals = {}
	 
	for i in range(len(cal_aips)):
		prgs.update(i) # Progress bar
		sol_time = obs_date + datetime.timedelta(float(cal_aips[i]['time']))
		time_unix = calendar.timegm(sol_time.timetuple())
		time_casa = time_unix + mod_jul_day
		if time_casa not in cals:
			cals[time_casa] = {}
		if DOBAND:
			time_interval = cal_aips[i]['interval']
		else:
			time_interval = cal_aips[i]['time_interval']
		time_interval *= 24 * 60 * 60 # Convert days to seconds
		field_id = cal_aips[i]['source_id'] - 1
		try:
			if DOBAND:
				antenna1 = ant_aips2ms(cal_aips[i]['antenna'])
			else:
				antenna1 = ant_aips2ms(cal_aips[i]['antenna_no'])
		except KeyError:
			# Ignore antennas not present in MS
			continue
		if antenna1 not in cals[time_casa]:
			cals[time_casa][antenna1] = {}
		for iif in range(ifs):
			if not iif in cals[time_casa][antenna1]:
				cals[time_casa][antenna1][iif] = {}
			row = iif * len(cal_aips) + i
			gain = numpy.empty(shape=(pols, chans), dtype='complex')
			if DOCALIB:
				mbdelay = numpy.empty(shape=(pols, chans), dtype='complex')
				sbdelay = numpy.empty(shape=(pols, chans), dtype='complex')
			snr = numpy.ones(shape=(pols,chans), dtype='float')
			for p in range(pols):
				if DOBAND:
					gain[p].real = cal_aips[i]['real_%d' % (p+1)][iif*chans:(iif+1)*chans]
					gain[p].imag = cal_aips[i]['imag_%d' % (p+1)][iif*chans:(iif+1)*chans]
				if ifs == 1:
					if DOCALIB:
						gain[p] = cal_aips[i]['real%d' % (p+1)] + 1j * cal_aips[i]['imag%d' % (p+1)]
						sbdelay[p] = cal_aips[i]['delay_%d' % (p+1)] * 1e9 # convert to nano-seconds
					snr[p] = cal_aips[i]['weight_%d' % (p+1)]
				else:
					if DOCALIB:
						gain[p] = cal_aips[i]['real%d' % (p+1)][iif] + 1j * cal_aips[i]['imag%d' % (p+1)][iif]
						sbdelay[p] = cal_aips[i]['delay_%d' % (p+1)][iif] * 1e9 # convert to nano-seconds
					snr[p] = cal_aips[i]['weight_%d' % (p+1)][iif]
				if DOCALIB:
					mbdelay[p] = cal_aips[i]['mbdelay%d' % (p+1)] * 1e9 # convert to nano-seconds
			zero_wt = numpy.logical_not(numpy.array(snr, dtype='bool'))
			# AIPS seems to use 3140 to indicate a bad solution
			flag_val = numpy.logical_and(numpy.array(gain.real, dtype='int') == 3140,
				   numpy.array(gain.imag, dtype='int') == 3140)
			flag = numpy.logical_or(zero_wt, flag_val)
			soln_ok = numpy.logical_not(flag)
			cals[time_casa][antenna1][iif] = {
				   'TIME': time_casa, 'INTERVAL': time_interval,
				   'SNR': snr, 'ANTENNA1': antenna1,
				   'CAL_DESC_ID': iif, 'FIELD_ID': field_id,
				   'SOLUTION_OK': soln_ok, 'FLAG': flag,
				   'GAIN': gain}
			if DOCALIB:
				cals[time_casa][antenna1][iif].update({'MBDELAY': mbdelay, 'SBDELAY': sbdelay})
	prgs.done()
	
	def fill_cal_desc(cal_table, ifs, chans, pols, ms_name):
		# Make an CAL_DESC entry for each AIPS IF
		cal_desc = table(cal_table + '/CAL_DESC', readonly=False)
		cal_desc.addrows(ifs)
		cal_desc.putcol('NUM_SPW', numpy.ones(shape=(ifs,), dtype='int'))
		cal_desc.putcol('NUM_CHAN', numpy.ones(shape=(chans,ifs), dtype='int'))
		cal_desc.putcol('NUM_RECEPTORS', numpy.array(ifs * [pols], dtype='int'))
		cal_desc.putcol('N_JONES', numpy.array(ifs * [2], dtype='int'))
		cal_desc.putcol('CHAN_FREQ', numpy.zeros(shape=(1,chans,ifs), dtype='float'))
		cal_desc.putcol('MEAS_FREQ_REF', numpy.zeros(shape=(ifs,), dtype='int'))
		cal_desc.putcol('CHAN_WIDTH', numpy.zeros(shape=(1,chans,ifs), dtype='float'))
		cal_desc.putcol('CHAN_RANGE', numpy.zeros(shape=(2,1,chans,ifs), dtype='int'))
		cal_desc.putcol('POLARIZATION_TYPE', numpy.zeros(shape=(1,chans,ifs), dtype='|S1'))
		cal_desc.putcol('JONES_TYPE', numpy.array(ifs * ['full']))
		cal_desc.putcol('MS_NAME', numpy.array(ifs * [os.path.basename(ms_name)]))
		cal_desc.putcol('SPECTRAL_WINDOW_ID', numpy.arange(ifs, dtype='int').reshape((1,ifs)))
		cal_desc.close()
	
	#if DOBAND:
	#	shutil.copytree('empty.bcal', cal_base + '.bcal')
	#	cal_table_dict = {'band': table(cal_base + '.bcal', readonly=False)}
	#else:
	#	shutil.copytree('empty.gcal', cal_base + '.gcal')
	#	shutil.copytree('empty.mbdcal', cal_base + '.mbdcal')
	#	shutil.copytree('empty.sbdcal', cal_base + '.sbdcal')
	#	cal_table_dict = {'gain': table(cal_base + '.gcal', readonly=False),
	#			  'mbdelay': table(cal_base + '.mbdcal', readonly=False),
	#			  'sbdelay': table(cal_base + '.sbdcal', readonly=False)}

	if DOBAND:
		empty_cal.create('bcal', cal_base)
		cal_table_dict = {'band': table(cal_base + '.bcal', readonly=False)}
	else:
		empty_cal.create('gcal', cal_base)
		empty_cal.create('mbdcal', cal_base)
		empty_cal.create('sbdcal', cal_base)
		cal_table_dict = {'gain': table(cal_base + '.gcal', readonly=False),
				  'mbdelay': table(cal_base + '.mbdcal', readonly=False),
				  'sbdelay': table(cal_base + '.sbdcal', readonly=False)}

	# Get the number of antennas in the MS. Casa seems to require a for
	# for each Time,Antenna,DataDesc.
	ms_ant = table(ms_name + '/ANTENNA')
	ms_num_ants = ms_ant.nrows()
	ms_ant.close()
	
	# Open tables and create required number of rows
	#if DOBAND:
	#	cal_table_dict = {'band': create_cal(cal_base + '.bcal', 'B')}
	#else:
	#	cal_table_dict = {
	#		'gain': create_cal(cal_base + '.gcal', 'G'),
	#		'mbdelay': create_cal(cal_base + '.mbdcal', 'KMBD'),
	#		'sbdelay': create_cal(cal_base + '.sbdcal', 'K')}
	
	# Initalise column values
	cal_rows = ms_num_ants * len(cals) * ifs
	total_solution_ok = numpy.ones(shape=(cal_rows,), dtype='bool')
	total_fit = numpy.ones(shape=(cal_rows,), dtype='float')
	total_fit_weight = numpy.ones(shape=(cal_rows,), dtype='float')
	fit = numpy.ones(shape=(1,chans,cal_rows), dtype='float')
	fit_weight = numpy.ones(shape=(1,chans,cal_rows), dtype='float')
	time = numpy.empty(shape=(ifs,ms_num_ants,len(cals)), dtype='float')
	time_list = cals.keys()
	time_list.sort()
	time[:,:,:] = time_list
	time = time.swapaxes(1,2).ravel()
	cal_desc_id = numpy.empty(shape=(ms_num_ants,len(cals),ifs), dtype='int')
	cal_desc_id[:,:,:] = range(ifs)
	cal_desc_id = cal_desc_id.swapaxes(0,2).ravel()
	antenna1 = numpy.empty(shape=(ifs,len(cals),ms_num_ants), dtype='float')
	antenna1[:,:,:] = range(ms_num_ants)
	antenna1 = antenna1.ravel()
	
	for cal_table in cal_table_dict.values():
		fill_cal_desc(cal_table.name(), ifs, chans, pols, ms_name)
		cal_table.addrows(cal_rows)
		cal_table.putcol('TIME', time)
		cal_table.putcol('ANTENNA1', antenna1)
		cal_table.putcol('CAL_DESC_ID', cal_desc_id)
		cal_table.putcol('TOTAL_SOLUTION_OK', total_solution_ok)
		cal_table.putcol('TOTAL_FIT', total_fit)
		cal_table.putcol('TOTAL_FIT_WEIGHT', total_fit_weight)
		cal_table.putcol('FIT', fit)
		cal_table.putcol('FIT_WEIGHT', fit_weight)
	
	del time, antenna1, cal_desc_id, total_solution_ok, total_fit, total_fit_weight, fit, fit_weight
	
	time_interval = numpy.empty(shape=(cal_rows,), dtype='float')
	field_id = numpy.empty(shape=(cal_rows,), dtype='int')
	snr = numpy.empty(shape=(pols,chans,cal_rows), dtype='float')
	solution_ok = numpy.empty(shape=(pols,chans,cal_rows), dtype='bool')
	flag = numpy.empty(shape=(pols,chans,cal_rows), dtype='bool')
	gain = numpy.empty(shape=(pols,chans,cal_rows), dtype='complex')
	if DOCALIB:
		mbdelay = numpy.empty(shape=(pols,chans,cal_rows), dtype='complex')
		sbdelay = numpy.empty(shape=(pols,chans,cal_rows), dtype='complex')
	
	timestamps = cals.keys()
	timestamps.sort()
	i = 0
	prgs = progress(cal_table_dict.values()[0].nrows())
	print 'Writing cal tables'
	for iif in range(ifs):
		for ts in timestamps:
			for ant in range(ms_num_ants):
				try:
					c = cals[ts][ant][iif]
					soln = True
				except KeyError:
					# If a solution does not exist for this ant/IF
					# create an standard one and mark it as bad.
					# Use another solution as a base.
					soln = False
				if soln:
					time_interval[i] = c['INTERVAL']
					field_id[i] = c['FIELD_ID']
					snr[:,:,i] = c['SNR']
					solution_ok[:,:,i] = c['SOLUTION_OK']
					flag[:,:,i] = c['FLAG']
					gain[:,:,i] = c['GAIN']
					if DOCALIB:
						mbdelay[:,:,i] = c['MBDELAY']
						sbdelay[:,:,i] = c['SBDELAY']
				else:
					dummy_ant = cals[ts].keys()[0]
					dummy_if = cals[ts][dummy_ant].keys()[0]
					time_interval[i] = cals[ts][dummy_ant][dummy_if]['INTERVAL']
					field_id[i] = cals[ts][dummy_ant][dummy_if]['FIELD_ID']
					snr[:,:,i] = 0
					solution_ok[:,:,i] = 0
					flag[:,:,i] = 1
					gain[:,:,i] = 1
					if DOCALIB:
						mbdelay[:,:,i] = 0
						sbdelay[:,:,i] = 0
				i += 1
				prgs.update(i)
	prgs.done()
	
	for cal_table in cal_table_dict.values():
		cal_table.putcol('SOLUTION_OK', solution_ok)
		cal_table.putcol('FLAG', flag)
		cal_table.putcol('GAIN', gain)
		cal_table.putcol('SNR', snr)
		cal_table.putcol('INTERVAL', time_interval)
		cal_table.putcol('FIELD_ID', field_id)
	
	if DOBAND:
		cal_table_dict['band'].putcol('GAIN', gain)
	else:
		cal_table_dict['gain'].putcol('GAIN', gain)
		cal_table_dict['mbdelay'].putcol('GAIN', mbdelay)
		cal_table_dict['sbdelay'].putcol('GAIN', sbdelay)
	
	for cal_table in cal_table_dict.values():
		cal_table.close()
	
	print 'Done.'

if __name__ == '__main__':
	casa_args_dbl = ['--rcdir', '--logfile']
	casa_args_sng = ['--maclogger', '--log2term', '--nologger', '--nogui', '--noipython', '-c']
	del sys.argv[0]
	for arg in casa_args_sng:
		if arg in sys.argv:
			sys.argv.remove(arg)
	for arg in casa_args_dbl:
		if arg in sys.argv:
			indx = sys.argv.index(arg)
			del sys.argv[indx:indx+2]
	print sys.argv

	if '-b' in sys.argv:
		cal_mode = 'BP'
		sys.argv.remove('-b')
	elif '-s' in sys.argv:
		cal_mode = 'SN'
		sys.argv.remove('-s')
	else:
		cal_mode = 'CL'

	if len(sys.argv) != 7 and len(sys.argv) != 8:
		print >> sys.stderr, 'Usage: %s <userno> <indisk> <cno> <inver> <ms> <outcal> [-b|-s] [antfile]' % sys.argv[0]
		sys.exit()
	userno, disk, cno, inver = [int(i) for i in sys.argv[1:5]]
	ms_name, cal_base = sys.argv[5:7]
	try:
		antfile = sys.argv[7]
	except IndexError:
		antfile = None

	importaipscal(ms_name, cal_base, userno, disk, cno, inver, cal_mode, antfile)

