#!/usr/bin/env ParselTongue
#
# Tool to convert AIPS Calibration tables to Measurement Set Calibration tables
# Stephen Bourke, JIVE
# Version 2010.12.13

import AIPS
from AIPSData import AIPSCat
from AIPSData import AIPSUVData
import Wizardry.AIPSData as wiz
from casa_utils import table
from progress import progress
import empty_cal
import sys
import shutil
import datetime
import calendar
import numpy
import os.path
from taskinit import casalog

mod_jul_day = 3506716800.0
sec_day = 86400.0 # Seconds per day

def importaipscal2(vis, caltable, userno, indisk, cno, inver, mode, antfile):
	if len(AIPS.disks) == 1:
		print >> sys.stderr, 'No AIPS disks. Use AIPSLite.attach_disk'
		return
	if mode.upper() not in ['CL', 'SN', 'BP']:
		print >> sys.stderr, "Invalid mode: '%s'. Must be 'CL', 'SN' or 'BP'." % mode
		return
	cal_mode = mode.upper()
	DOBAND = cal_mode == 'BP'
	DOCALIB = not DOBAND

	AIPS.userno = userno
	ms_name = vis
	cal_base = caltable


	# Select dataset and table
	ds = [i for i in AIPSCat()[indisk] if i['cno'] == cno][0]
	uv = wiz.AIPSUVData(ds['name'], ds['klass'], indisk, ds['seq'])
	obs_date = datetime.datetime(*[int(i) for i in uv.header.date_obs.split('-')])
	#print AIPSUVData(uv), '%s%d' % (mode, inver), '->', ms_name, cal_base
	casalog.post('Reading AIPS dataset "Disk %d, %s.%s.%d"' % (uv.disk, uv.name, uv.klass, uv.seq))
	casalog.post('for MS %s' % ms_name)	
	
	cal_aips = uv.table(mode, inver)
	pols = cal_aips.keywords['NO_POL']
	ifs = cal_aips.keywords['NO_IF']	# AIPS IF / Casa SPW
	if DOBAND:
		chans = cal_aips.keywords['NO_CHAN']
	else:
		chans = 1
	
	# User specified mapping from AIPS ants to MS ant ids
	ant_mapping = {}
	if antfile:
		inf = open(antfile)
		for line in inf:
			aips_ant_no, ms_ant_id = [int(x) for x in line.split()]
			ant_mapping[aips_ant_no] = ms_ant_id
		inf.close()
	
	def ant_aips2ms(ant_num):
		if ant_mapping:
			return ant_mapping[ant_num]
		else:
			return ant_num - 1
	
	# Create record type, and array to store Cal values.
	cal_dtype = [('TIME', 'float'), ('INTERVAL', 'float'), ('SNR', 'float', (pols,chans)),
		     ('ANTENNA1', 'int'), ('CAL_DESC_ID', 'int'), ('FIELD_ID', 'int'),
		     ('SOLUTION_OK', 'bool', (pols,chans)), ('FLAG', 'bool', (pols,chans)),
		     ('GAIN', 'complex', (pols,chans)), ('MBDELAY', 'complex', (pols,chans)),
		     ('SBDELAY', 'complex', (pols,chans))]
	
	cals = numpy.empty(shape=(len(cal_aips)*ifs,), dtype=cal_dtype)
	
	sys.stdout.write('Reading %s table\n' % mode)
	sys.stdout.flush()
	
	# Populate the calibration array from the AIPS table
	prgs = progress(len(cal_aips) * ifs)
	i = 0
	for row in cal_aips:
		prgs.update(i) # Progress bar
		sol_time = obs_date + datetime.timedelta(float(row['time']))
		time_unix = calendar.timegm(sol_time.timetuple())
		time_casa = time_unix + mod_jul_day
		if DOBAND:
			interval = row['interval'] * sec_day
		else:
			interval = row['time_interval'] * sec_day
		try:
			if DOBAND:
				antenna1 = ant_aips2ms(row['antenna'])
			else:
				antenna1 = ant_aips2ms(row['antenna_no'])
		except KeyError:
			# Ignore antennas not present in MS
			continue
		for iif in range(ifs):
			cals[i]['TIME'] = time_casa
			cals[i]['INTERVAL'] = interval
			cals[i]['FIELD_ID'] = row['source_id'] - 1
			cals[i]['ANTENNA1'] = antenna1
			cals[i]['CAL_DESC_ID'] = iif
			for p in range(pols):
				if DOBAND:
					cals[i]['GAIN'][p,:].real = row['real_%d' % (p+1)][iif*chans:(iif+1)*chans]
					cals[i]['GAIN'][p,:].imag = row['imag_%d' % (p+1)][iif*chans:(iif+1)*chans]
					cals[i]['SBDELAY'][p,:] = 0
					cals[i]['MBDELAY'][p,:] = 0
				if ifs == 1:
					if DOCALIB:
						cals[i]['GAIN'][p,:] = row['real%d' % (p+1)] + 1j * row['imag%d' % (p+1)]
						cals[i]['SBDELAY'][p,:] = row['delay_%d' % (p+1)] * 1e9 # convert to nano-seconds
					cals[i]['SNR'][p,:] = row['weight_%d' % (p+1)]
				else:
					if DOCALIB:
						cals[i]['GAIN'][p,:] = row['real%d' % (p+1)][iif] + 1j * row['imag%d' % (p+1)][iif]
						cals[i]['SBDELAY'][p,:] = row['delay_%d' % (p+1)][iif] * 1e9 # convert to nano-seconds
					cals[i]['SNR'][p,:] = row['weight_%d' % (p+1)][iif]
				if DOCALIB:
					cals[i]['MBDELAY'][p,:] = row['mbdelay%d' % (p+1)] * 1e9 # convert to nano-seconds
			zero_wt = numpy.logical_not(cals[i]['SNR'])
			# AIPS seems to use 3140 to indicate a bad solution
			flag_val = numpy.logical_and(numpy.int_(cals[i]['GAIN'].real) == 3140,
				   		     numpy.int_(cals[i]['GAIN'].imag) == 3140)
			cals[i]['FLAG'][:] = numpy.logical_or(zero_wt, flag_val)
			cals[i]['SOLUTION_OK'][:] = numpy.logical_not(cals[i]['FLAG'])
			i += 1
	prgs.done()
	cals.resize((i,))
	cals.sort(order=['CAL_DESC_ID','TIME','ANTENNA1'])

	import pickle
	outf = open('/tmp/calin.pkl', 'w')
	pickle.dump(cals, outf)
	outf.close()
	
	def fill_cal_desc(cal_table, ifs, chans, pols, ms_name):
		# Make an CAL_DESC entry for each AIPS IF
		cal_desc = table(cal_table + '/CAL_DESC', readonly=False)
		cal_desc.addrows(ifs)
		cal_desc.putcol('NUM_SPW', numpy.ones(shape=(ifs,), dtype='int'))
		cal_desc.putcol('NUM_CHAN', numpy.ones(shape=(chans,ifs), dtype='int'))
		cal_desc.putcol('NUM_RECEPTORS', numpy.array(ifs * [pols], dtype='int'))
		cal_desc.putcol('N_JONES', numpy.array(ifs * [2], dtype='int'))
		cal_desc.putcol('CHAN_FREQ', numpy.zeros(shape=(1,chans,ifs), dtype='float'))
		cal_desc.putcol('MEAS_FREQ_REF', numpy.zeros(shape=(ifs,), dtype='int'))
		cal_desc.putcol('CHAN_WIDTH', numpy.zeros(shape=(1,chans,ifs), dtype='float'))
		cal_desc.putcol('CHAN_RANGE', numpy.zeros(shape=(2,1,chans,ifs), dtype='int'))
		cal_desc.putcol('POLARIZATION_TYPE', numpy.zeros(shape=(1,chans,ifs), dtype='|S1'))
		cal_desc.putcol('JONES_TYPE', numpy.array(ifs * ['full']))
		cal_desc.putcol('MS_NAME', numpy.array(ifs * [os.path.basename(ms_name)]))
		cal_desc.putcol('SPECTRAL_WINDOW_ID', numpy.arange(ifs, dtype='int').reshape((1,ifs)))
		cal_desc.close()
	
	# Get the number of antennas in the MS. Casa seems to require a for
	# for each Time,Antenna,DataDesc.
	ms_ant = table(ms_name + '/ANTENNA')
	ms_num_ants = ms_ant.nrows()
	ms_ant.close()
	
	timestamps = numpy.unique(cals['TIME'])
	cal_rows = ms_num_ants * len(timestamps) * ifs
	
	# Open tables and create required number of rows
	if DOBAND:
		empty_cal.create('bcal', cal_base)
		cal_table_dict = {'band': table(cal_base + '.bcal', readonly=False)}
	else:
		empty_cal.create('gcal', cal_base)
		empty_cal.create('mbdcal', cal_base)
		empty_cal.create('sbdcal', cal_base)
		cal_table_dict = {'gain': table(cal_base + '.gcal', readonly=False),
				  'mbdelay': table(cal_base + '.mbdcal', readonly=False),
				  'sbdelay': table(cal_base + '.sbdcal', readonly=False)}
	
	# Initalise column values that do not depend on solutions
	total_solution_ok = numpy.ones(shape=(cal_rows,), dtype='bool')
	total_fit = numpy.ones(shape=(cal_rows,), dtype='float')
	total_fit_weight = numpy.ones(shape=(cal_rows,), dtype='float')
	fit = numpy.ones(shape=(1,chans,cal_rows), dtype='float')
	fit_weight = numpy.ones(shape=(1,chans,cal_rows), dtype='float')
	
	for cal_table in cal_table_dict.values():
		fill_cal_desc(cal_table.name(), ifs, chans, pols, ms_name)
		cal_table.addrows(cal_rows)
		cal_table.putcol('TOTAL_SOLUTION_OK', total_solution_ok)
		cal_table.putcol('TOTAL_FIT', total_fit)
		cal_table.putcol('TOTAL_FIT_WEIGHT', total_fit_weight)
		cal_table.putcol('FIT', fit)
		cal_table.putcol('FIT_WEIGHT', fit_weight)
	
	del total_solution_ok, total_fit, total_fit_weight, fit, fit_weight
	
	# Casa seems to expect a calibration row for each ant/cal_desc/time
	# Create an populate an output array filling rows for which a solution
	# doesn't exist with a flagged solution.
	cal_out = numpy.empty(shape=(cal_rows,), dtype=cal_dtype)
	prgs = progress(cal_rows)
	
	casalog.post('Creating output columns.')
	i = 0
	cal_out[0] = cals[0]
	for iif in range(ifs):
		for time_indx in range(len(timestamps)):
			for ant in range(ms_num_ants):
				out_indx = iif * len(timestamps) * ms_num_ants + \
					time_indx * ms_num_ants + ant
				if cals[i]['CAL_DESC_ID'] == iif and \
				   cals[i]['TIME'] == timestamps[time_indx] and \
				   cals[i]['ANTENNA1'] == ant:
					cal_out[out_indx] = cals[i]
					i += 1
				else:
					# FIXME: previous record may have wrong field_id, etc.
					cal_out[out_indx] = cal_out[out_indx-1]
					cal_out[out_indx]['TIME'] = timestamps[time_indx]
					cal_out[out_indx]['CAL_DESC_ID'] = iif
					cal_out[out_indx]['ANTENNA1'] = ant
					cal_out[out_indx]['SNR'][:] = 0
					cal_out[out_indx]['SOLUTION_OK'][:] = 0
					cal_out[out_indx]['FLAG'][:] = 1
					cal_out[out_indx]['GAIN'][:] = 1
					if DOCALIB:
						cal_out[out_indx]['MBDELAY'][:] = 0
						cal_out[out_indx]['SBDELAY'][:] = 0
				prgs.update()
	prgs.done()

	outf = open('/tmp/calout.pkl', 'w')
	pickle.dump(cal_out, outf)
	outf.close()	
	
	def rotateaxes(arr):
		for i in range(len(arr.shape)-1):
			arr = arr.swapaxes(i, i+1)
		return arr
	
	casalog.post('Writing cal tables.')
	for cal_table in cal_table_dict.values():
		cols = list(cal_out.dtype.names)
		cols.remove('GAIN')
		cols.remove('MBDELAY')
		cols.remove('SBDELAY')
		for col_name in cols:
			# Col doesn't always get set properly unless a new array is initalised.
			cal_table.putcol(col_name, numpy.array(rotateaxes(cal_out[col_name])))
	
	if DOBAND:
		cal_table_dict['band'].putcol('GAIN', numpy.array(rotateaxes(cal_out['GAIN'])))
	else:
		cal_table_dict['gain'].putcol('GAIN', numpy.array(rotateaxes(cal_out['GAIN'])))
		cal_table_dict['mbdelay'].putcol('GAIN', numpy.array(rotateaxes(cal_out['MBDELAY'])))
		cal_table_dict['sbdelay'].putcol('GAIN', numpy.array(rotateaxes(cal_out['MBDELAY'])))
	
	for cal_table in cal_table_dict.values():
		cal_table.close()
	
if __name__ == '__main__':
	casa_args_dbl = ['--rcdir', '--logfile']
	casa_args_sng = ['--maclogger', '--log2term', '--nologger', '--nogui', '--noipython', '-c']
	del sys.argv[0]
	for arg in casa_args_sng:
		if arg in sys.argv:
			sys.argv.remove(arg)
	for arg in casa_args_dbl:
		if arg in sys.argv:
			indx = sys.argv.index(arg)
			del sys.argv[indx:indx+2]
	print sys.argv

	if '-b' in sys.argv:
		cal_mode = 'BP'
		sys.argv.remove('-b')
	elif '-s' in sys.argv:
		cal_mode = 'SN'
		sys.argv.remove('-s')
	else:
		cal_mode = 'CL'

	if len(sys.argv) != 7 and len(sys.argv) != 8:
		print >> sys.stderr, 'Usage: %s <userno> <indisk> <cno> <inver> <ms> <outcal> [-b|-s] [antfile]' % sys.argv[0]
		sys.exit()
	userno, disk, cno, inver = [int(i) for i in sys.argv[1:5]]
	ms_name, cal_base = sys.argv[5:7]
	try:
		antfile = sys.argv[7]
	except IndexError:
		antfile = None

	importaipscal2(ms_name, cal_base, userno, disk, cno, inver, cal_mode, antfile)

