#!/usr/bin/env python
#
# Program to geometrically shift a field in a Measurement Set
# Stephen Bourke, jive
# Dec, 2010

import math
import numpy
import cmath
import os.path
import progress
from taskinit import casalog, qa
from casa_utils import table, get_field_id

C = 299792458.0

def shift_pos(ra, dec, rashift, decshift):
	return ra + rashift / math.cos(dec), dec + decshift

def shift(msname, field, shiftval, nrows=2**10):
	"""Geometrically shift a field in a Measurement Set.

	'field' can be field ID or name.
	'shiftval' is a RA, DEC measure, eg: ['1arcsec', '1arcsec']
	'nrows' is the number of rows processed per loop,
	 must fit in memory, defaults to 1024."""

	casalog.origin('shift')
	field_id = get_field_id(msname, field)
	rashift = qa.convert(shiftval[0], 'rad')['value']
	if len(shiftval) == 1:
		decshift = 0
	else:
		decshift = qa.convert(shiftval[1], 'rad')['value']
	if rashift == 0 and decshift == 0:
		casalog.post('No shiftval set.')
		return

	cos_rashift = math.cos(rashift)
	sin_rashift = math.sin(rashift)
	cos_decshift = math.cos(decshift)
	sin_decshift = math.sin(decshift)
		
	# Wavelengths
	casalog.post('Getting wavelengths.')
	sp_win = table(msname + '/SPECTRAL_WINDOW')
	wl = C / sp_win.getcol('CHAN_FREQ')
	sp_win.close()
	
	# DataDescID to SPW mapping
	data_desc = table(msname + '/DATA_DESCRIPTION')
	desc2spw = data_desc.getcol('SPECTRAL_WINDOW_ID')
	data_desc.close()
	
	# Rotate UVWs and update phase
	casalog.post('Selecting data.')
	main_tbl = table(msname, readonly=False)
	subt = main_tbl.query('FIELD_ID==%d' % field_id, columns='UVW, DATA_DESC_ID, DATA', sortlist='TIME')
	startrow = 0
	uvw = subt.getcol('UVW', startrow, nrows)
	ddsc = subt.getcol('DATA_DESC_ID', startrow, nrows)
	vis_data = subt.getcol('DATA', startrow, nrows)
	casalog.post('Rotating UVWs and shifting phases.')
	shift_progress = progress.progress(subt.nrows())
	for i in range(subt.nrows()):
		if i >= startrow + nrows:
			# Write back data and get next set of rows
			subt.putcol('UVW', uvw, startrow, nrows)
			subt.putcol('DATA', vis_data, startrow, nrows)
			startrow += nrows
			uvw = subt.getcol('UVW', startrow, nrows)
			ddsc = subt.getcol('DATA_DESC_ID', startrow, nrows)
			vis_data = subt.getcol('DATA', startrow, nrows)
		indx = i % nrows
		u0,v0,w0 = uvw[:,indx]
		u1 = u0 * cos_rashift - w0 * sin_rashift
		wT = u0 * sin_rashift + w0 * cos_rashift
		v1 = v0 * cos_decshift - wT * sin_decshift
		w1 = v0 * sin_decshift + wT * cos_decshift
		uvw[:,indx] = [u1, v1, w1]
		# Shift the visibility phase based on the delay (w) difference
		spw = desc2spw[ddsc[indx]]
		delay_angle = (w1 - w0) / wl[:,spw] * -2 * math.pi
		complex_delay = numpy.exp(1j * delay_angle)
		vis = vis_data[:,:,indx]
		for corr in range(vis.shape[0]):
			vis[corr,:] *= complex_delay
		vis_data[:,:,indx] = vis
		shift_progress.update(i)
	# FIXME: Confirm that it's ok for data size to be less than nrows
	subt.putcol('UVW', uvw, startrow, nrows)
	subt.putcol('DATA', vis_data, startrow, nrows)
	shift_progress.done()
	subt.close()
	main_tbl.close()
	
	# Update position in FIELD subtable
	casalog.post('Updating positions in FIELD table.')
	field_tbl = table(msname + '/FIELD', readonly=False)
	if field_id < field_tbl.nrows():
		dir_cell = field_tbl.getcell('PHASE_DIR', field_id)
		ra, dec = dir_cell[:,0]
		newpos = shift_pos(ra, dec, rashift, decshift)
		dir_cell[:,0] = newpos
		field_tbl.putcell('PHASE_DIR', field_id, dir_cell)
	field_tbl.close()
