# Functions to get various attributes of AIPS datasets.
# Stephen Bourke, NUI Galway, JIVE.

import gen_utils, unit_utils
import math

def frequency(dataset):
	"""Returns the Frequency of the observation in Hz."""
	indx = dataset.header['ctype'].index('FREQ')
	return dataset.header['crval'][indx]

def wavelength(dataset):
	"""Returns the observing wavelength in metres."""
	return unit_utils.freq2wl(frequency(dataset))

def subbands(dataset):
	"""Returns the number of subbands (IF's in AIPS.)"""
	indx = dataset.header['ctype'].index('IF')
	return dataset.header['naxis'][indx]

def channels(dataset):
	"""Returns the number of channels per subband."""
	indx = dataset.header['ctype'].index('FREQ')
	return dataset.header['naxis'][indx]

def channel_width(dataset):
	"""Returns the channel width."""
	indx = dataset.header['ctype'].index('FREQ')
	return dataset.header['cdelt'][indx]

def channel_freq(dataset, channel=1):
	"""Returns the frequency of a specific channel in a dataset."""
	indx = dataset.header['ctype'].index('FREQ')
	channel_0 = dataset.header['crpix'][indx]
	return frequency(dataset) + \
		(channel - channel_0) * channel_width(dataset)

def freq2chan(dataset, freq):
	"""Returns the channel number that corrisponds to input frequency."""
	indx = dataset.header['ctype'].index('FREQ')
	channel_0 = dataset.header['crpix'][indx]
	channel_offset = (freq - frequency(dataset)) / channel_width(dataset)
	return int(round(channel_0 + channel_offset))

def ra(dataset):
	"""Returns the RA specified in the header.
	May not be useful with multi-source files."""
	#Divide by 15 to convert from degrees to hours
	return ra_deg(dataset) / 15

def ra_deg(dataset):
	"""Returns the RA in degrees as specified in the header.
	May not be useful with multi-source files."""
	try:
		indx = dataset.header['ctype'].index('RA')
	except ValueError:
		indx = dataset.header['ctype'].index('RA---SIN')
	ra_deg = dataset.header['crval'][indx]
	if ra_deg < 0:
		ra_deg = ra_deg + 360
	return ra_deg

def dec(dataset):
	"""Returns the DEC specified in the header.
	May not be useful with multi-source files."""
	try:
		indx = dataset.header['ctype'].index('DEC')
	except ValueError:
		indx = dataset.header['ctype'].index('DEC--SIN')
	return dataset.header['crval'][indx]

def is_multi(dataset):
	"""Return True if argument is a multi-source dataset."""
	return 'SOURCE' in dataset.header['ptype']

def subband_bandwidth(dataset):
	"""Return the subband bandwidth of the observation."""
	return channels(dataset) * channel_width(dataset)

def total_bandwidth(dataset):
	"""Return the total bandwidth of the observation."""
	return subbands(dataset) * subband_bandwidth(dataset)

def antennas(dataset):
	"""Returns the number of antennas in an observation."""
	num_ants = 0
	ant_table = dataset.table('AN', 1)
	for ant in ant_table:
	 	# VLA antennas that are 'out' have locations of 0,0,0
		if ant.stabxyz != [0.0, 0.0, 0.0]:
			num_ants += 1
	return num_ants
	
def ant_names(dataset):
	"""Returns the number of antennas in an observation."""
	ant_table = dataset.table('AN', 1)
	names = []
	for ant in ant_table:
		names.append(ant['anname'].strip())
	return names

def antenna_coords(dataset):
	"""Returns a list of XYZ locations of the antennas relative to
	the reference position, in metres."""
	antenna_coords = []
	ant_table = dataset.table('AN', 1)
	for ant in ant_table:
	 	# VLA antennas that are 'out' have locations of 0,0,0
		if ant.stabxyz != [0.0, 0.0, 0.0]:
			antenna_coords.append(ant.stabxyz)
	return antenna_coords

def max_baseline(dataset):
	"""Calculate the maximum baseline length (ie. If the source was at the
	zenith of that baseline.)
	Result is in metres."""
	ant_coords = antenna_coords(dataset)
	max_baseline = 0
	for ant1 in range(len(ant_coords) - 1):
		for ant2 in range(ant1 + 1, len(ant_coords)):
			x = ant_coords[ant1][0] - ant_coords[ant2][0]
			y = ant_coords[ant1][1] - ant_coords[ant2][1]
			z = ant_coords[ant1][2] - ant_coords[ant2][2]
			baseline_length = math.sqrt(x*x + y*y + z*z)
			if baseline_length > max_baseline:
				max_baseline = baseline_length
	return max_baseline

def resolution_asec(dataset):
	"""Returns the resoultion in arc-seconds."""
	resolution = wavelength(dataset) / max_baseline(dataset)
	#convert to arc_seconds
	return math.degrees(resolution) * 3600

def wiz_int_time(dataset, iterations=100):
	"""Return the integration time for the obsevation in seconds.
	
	AIPS allows random integration times within an observation. The
	integration time is normally constant, however if the dataset has been
	time-averaged by UVAVG, the integration times might be slightly erratic.
	This method calculates t_int by looking at the first 100 integrations
	on the first baseline and taking the median value as t_int. It should
	cope reasonably well with UVAVG'd data."""
	
	try:
		dataset.__iter__()
	except AttributeError:
		raise TypeError, 'Wizardy dataset required.'
	
	iter = 0
	t_ints = []
	for vis in dataset:
		t_curr = vis.time
		if iter == 0:
			t_0 = t_prev = t_curr
			ref_baseline = vis.baseline
		if vis.baseline != ref_baseline:
			# Look only at a single baseline
			continue
		if iter > iterations:
			break
		t_ints.append(t_curr - t_prev)
		t_prev = t_curr
		iter += 1
	# Get median value and convert from days to seconds
	return gen_utils.median(t_ints) * 24*60*60

def bw_fov_asec(dataset):
	"""Returns the bandwidth limited field of view. (90% Amp.)"""
	#Based on formula from EVN user guide.
	return (1e12 * 49.5 * channels(dataset)) / (max_baseline(dataset) * subband_bandwidth(dataset))

def wiz_time_fov_asec(dataset):
	"""Returns the time limited field of view. (90% Amp.)"""
	#Based on formula from EVN user guide.
	return 1e8 * 18.56 * wavelength(dataset) / (max_baseline(dataset) * wiz_int_time(dataset))

def wiz_fov_asec(dataset):
	"""Returns the imageable Field of View.
	(Where amplitudes will have 90% of their value at the edge due to
	bandwidth and time smearing."""
	return min(bw_fov_asec(dataset), wiz_time_fov_asec(dataset))

def int_time_for_fov_asec(dataset, fov):
	"""Returns the integration time required to allow the desired
	field of view."""
	return 1e8 * 18.56 * wavelength(dataset) / (max_baseline(dataset) * fov)

def channels_for_fov_asec(dataset, fov):
	"""Returns the number of channels necessary to allow the
	desired field of view."""
	return math.ceil((fov * subband_bandwidth(dataset) * max_baseline(dataset)) / (1e12 * 49.5))

def channel_width_for_fov_asec(dataset, fov):
	"""Returns the channel width necessary to allow the desired field of view."""
	return (1e12 * 49.5) / (fov * max_baseline(dataset))

def imsize_pix(dataset):
	"""Returns the Image dimensions in pixels."""
	indx = dataset.header['ctype'].index('RA---SIN')
	dim_x = dataset.header['naxis'][indx]
	indx = dataset.header['ctype'].index('DEC--SIN')
	dim_y = dataset.header['naxis'][indx]
	return dim_x, dim_y

def ref_pixel(dataset):
	"""Return the location of the reference pixel (in pixels.)"""
	indx = dataset.header['ctype'].index('RA---SIN')
	ref_x = dataset.header['crpix'][indx]
	indx = dataset.header['ctype'].index('DEC--SIN')
	ref_y = dataset.header['crpix'][indx]
	return ref_x, ref_y

def pixel_inc_deg(dataset):
	"""Returns the cellsize (pixel size) in degrees."""
	indx = dataset.header['ctype'].index('RA---SIN')
	inc_x = dataset.header['cdelt'][indx]
	indy = dataset.header['ctype'].index('DEC--SIN')
	inc_y = dataset.header['cdelt'][indx]
	return inc_x, inc_y

def pixel_inc_asec(dataset):
	"""Returns the cellsize (pixel size) in arc-seconds."""
	return [3600 * i for i in pixel_inc_deg(dataset)]
