# Utility functions for dealing with AIPS catalogs & disks.
# Stephen Bourke, NUI Galway, JIVE.

import AIPS
from AIPSData import AIPSCat, AIPSUVData, AIPSImage
from AIPSTask import AIPSTask
import gen_utils
import re, os


def getname(cno, disk=1):
    cat = AIPSCat(disk)[disk]
    results = [entry for entry in cat if entry.cno == cno]
    if len(results) == 0:
        raise IndexError, 'Entry %d not found' % cno
    if len(results) > 1:
        raise RuntimeError, 'Multiple matches found. This should not happen!'
    else:
        dataset = results[0]
        if dataset.type == 'MA':
            return AIPSImage(dataset.name, dataset.klass, disk, dataset.seq)
        elif dataset.type == 'UV':
            return AIPSUVData(dataset.name, dataset.klass, disk, dataset.seq)
        else:
            raise TypeError, 'Not an Image or UV dataset'


def disk_with_file(name, klass, seq, cno):
    """Find the disk on which the dataset specified resides.
    
    Raise RuntimeException if it is not found or found
    on more than one disk."""
    
    matches = []
    cat = AIPSCat()
    for disk in cat:
        for entry in cat[disk]:
            if entry.name == name and entry.klass == klass and \
               entry.seq == seq and entry.cno == cno:
                matches.append(disk)
    if len(matches) > 1:
        raise RuntimeError, 'Entry found on more than one disk'
    elif len(matches) == 0:
        raise IndexError, 'Dataset not found'
    else:
        return matches[0]


def is_remote_disk(disknum):
    if AIPS.disks[disknum].url:
        return True
    else:
        return False

def host_with_disk(disknum):
    """Returns the host on which disknum is located"""
    if is_remote_disk(disknum):
        host = re.findall('^http://(.*):\d+/$', AIPS.disks[disknum].url)[0]
    else:
        host = 'localhost'
    return host


def disk_index(url, native_index):
    """Returns the disk ParselTongue id of disk x on host y if it is one
    of our disks."""
    found = []
    for i in range(1, len(AIPS.disks)):
        disk = AIPS.disks[i]
        if disk.url == url and disk.disk == native_index:
           found.append(i)
    #for i in range(1, len(AIPS.disks)):
    #    if host_with_disk(i) == host and AIPS.disks[i].disk == native_index:
    #       found.append(i)
    if len(found) < 1:
        raise IndexError
    elif len(found) > 1:
        raise RuntimeError, 'Disk is mapped multiple times: %s' % str(found)
    else:
        return found[0]
            

def out_files(task):
    """Tries to deduce the files created by last invocation of task object.
    
    It does this by looking for lines in the tasks output of the form:
    TSKNM1: Create NAME        .CLASS .   1 (MA)  on disk  1  cno    1
    It will not work if the task doesn't produce such messages."""
    
    # Regex to extract name, class, sequence, type, and catalog number from a line as above.
    pattern = re.compile('.*Create (.{12})\.(.{6})\.\s*(\d+)\s*\((..)\)\s*on disk\s*(\d+)\s*cno\s*(\d+)')
    
    outfiles = []
    for line in task.messages():
        match = pattern.match(line)
        if match:
            name, klass, seq, typ, disk_on_host, cno = match.groups()
            name, klass = [i.rstrip(' ') for i in (name, klass)]
            seq, disk_on_host, cno = [int(i) for i in (seq, disk_on_host, cno)]
            # This is the number of the disk as seen by the AIPS task and can
            # be (and probably will be) different the ParselTongue number for
            # the disk.
            
            # There is a bug here, out2disk could translate to
            # the same number as outdisk
            #task_outdisk = int(task.outdisk)
            #if AIPS.disks[task_outdisk].disk == disk_on_host:
            #    disk = task_outdisk
            
            if typ == 'UV' or typ == 'MA':
                try:
		    # This can be slow if remote disks are involved
		    disk = disk_with_file(name, klass, seq, cno)
                except IndexError:
		    # Scratch file?
		    pass
		else:
                    if typ == 'UV':
                        dataset = AIPSUVData(name, klass, disk, seq)
                    else:
                        dataset = AIPSImage(name, klass, disk, seq)
                    outfiles.append(dataset)
    return outfiles


def check_disk(disk):
    """Return True or False based on an access test to the disk."""
    
    AIPS.disks.append(disk)
    try:
        # If AIPSCat doesn't throw an exception, assume the disk is ok.
        AIPSCat(len(AIPS.disks)-1)
    except:
        disk_ok=False
    else:
        disk_ok=True
    AIPS.disks.pop()
    return disk_ok


def copy_data(dataset, outdisk, tempfile=None, shared_area=None):
    """Copy the dataset to the selected disk.
    
    Allows data to be copied from when disks are not
    local. Requires a shared disk.
    Returns a list containing the new dataset(s)."""
    
    # TODO: If disks are on same machine use MOVE.
    
    fittp = AIPSTask('fittp')
    fittp.indata = dataset
    if not tempfile:
        if not shared_area:
            # If no shared area is specified assume
            # home dir is shared.
            shared_area = os.environ['HOME']
        tempfile = gen_utils.rand_string()
        tempfile = os.path.join(shared_area, tempfile)
        assert not os.path.exists(tempfile)
    fittp.outfile = tempfile
    fittp()
    
    fitld = AIPSTask('fitld')
    fitld.infile = tempfile
    fitld.outdisk = outdisk
    fitld()
    
    os.remove(tempfile)
    return out_files(fitld)


def wipe_disk(disknum):
    """Wipes all the current user's data from the disk."""
    for entry in AIPSCat(disknum)[disknum]:
        if entry.type == 'MA':
            file = AIPSImage(entry.name, entry.klass, disknum, entry.seq)
        elif entry.type == 'UV':
            file = AIPSUVData(entry.name, entry.klass, disknum, entry.seq)
        else:
            print >> sys.stderr, 'Could not remove', entry
            continue
        file.zap()


def load_remote_disks(diskfile):
    """Loads disks from file consisting of:
    /path/doesnt/matter url disk"""
    disks_added = 0
    srv_fl = open(diskfile)
    for line in srv_fl:
        path, url, disk = line.split()
        AIPS.disks.append(AIPS.AIPSDisk(url, int(disk)))
        disks_added += 1
    srv_fl.close()
    return disks_added
