# Functions to deal with / convert units.
# Stephen Bourke, NUI Galway, JIVE.

import math

C = 299792458.0

def dhms2sec(aips_tm):
	"""Convert time in [dd, hh, mm, ss.s] to time in seconds."""
	d = aips_tm[0] * 24 * 60 * 60
	h = aips_tm[1] * 60 * 60
	m = aips_tm[2] * 60
	s = aips_tm[3]
	return d + h + m + s

def sec2dhms(s):
	"""Convert time in seconds to (dd, hh, mm, ss.s)"""
	hours, days = math.modf(float(s) / (24 * 60 * 60))
	hours = hours * 24
	minutes, hours = math.modf(hours)
	minutes = minutes * 60
	seconds, minutes = math.modf(minutes)
	seconds = seconds * 60
	return days, hours, minutes, seconds

def deg2dms(angle):
	"""Convert an angle in degrees to (degrees, minutes, seconds)"""
	minutes, degrees = math.modf(angle)
	minutes = abs(minutes * 60)
	seconds, minutes = math.modf(minutes)
	seconds = seconds * 60
	return degrees, minutes, seconds

def dms2deg(dms):
	"""Convert an angle in (degrees, minutes, seconds) to degrees."""
	return dms[0] + float(dms[1]) / 60 + float(dms[2]) / 3600

def dms2str(dms, sep=':', precision=6, sign=False):
	"""Take a [dd, mm, ss] sequence and return a string representation.
	
	precision is the number of decimal places required for the
	(arc)seconds part. sign determins wither the string will
	start with a +/- character."""
	
	# Construct the %f part for the (arc)seconds specifier of the
	# format string.
	if precision < 0:
		raise ValueError, 'precision cannot be negative'
	elif precision == 0:
		float_spec = '%02.0f'
	else:
		float_spec = '%0' + str(precision+3) + '.' + str(precision) + 'f'

	if sign and dms[0] >= 0:
		sign = '+'
	else:
		# Don't print a '-' It will be printed automatically.
		sign = ''
	
	return ('%s%02d%s%02d%s' + float_spec) % \
		(sign, int(dms[0]), sep, int(dms[1]), sep, dms[2])

def deg2str(angle, sep=':', precision=6, sign=False):
	"""Takes a decimal degree input and formats it as a dd:mm:ss.s string."""
	return dms2str(deg2dms(angle), sep, precision, sign)

def str2dms(angle, sep=':'):
	d,m,s = [float(i) for i in angle.split(sep)]
	if d < 0:
		m *= -1
		s *= -1
	return (d,m,s)

def str2deg(angle, sep=':'):
	return dms2deg(str2dms(angle, sep))

def freq2wl(freq):
	"""Convert frequency to wavelength."""
	return C / freq

def wl2freq(wl):
	"""Convert wavelength to frequency."""
	return C / wl

def freq2str(f, precision=3):
	"""Returns a 'pretty' string for an input frequency in Hz."""
	if f > 1e9:
		return ('%.'+str(precision)+'f GHz') % (f / 1e9)
	elif f > 1e6:
		return ('%.'+str(precision)+'f MHz') % (f / 1e6)
	elif f > 1000:
		return ('%.'+str(precision)+'f KHz') % (f / 1000)
	else:
		return ('%.'+str(precision)+'f Hz') % f

def length2str(l, precision=3):
	"""Returns a 'pretty' string for an input length in Metres."""
	if l > 1000:
		return ('%.'+str(precision)+'f km') % (l / 1000)
	elif l > 1:
		return ('%.'+str(precision)+'f m') % l
	elif l > 0.1:
		return ('%.'+str(precision)+'f cm') % (l * 100)
	else:
		return ('%.'+str(precision)+'f mm') % (l * 1000)
