#!/usr/bin/env python

# Copyright (C) 2009
# The MeqTree Foundation & 
# JIVE (Joint Institute for VLBI in Europe)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# Create an Measurement Set Antenna table using the values from a text file.
# Stephen Bourke, JIVE.
# Sept, 2009-.
#
# Known Arrays:
#  EVN EEVN
#  ELOFAR_LBA ELOFAR_HBA
#  GLOBAL
#  GMRT
#  GMVA
#  KAT7
#  LOFAR_CORE_LBA LOFAR_CORE_HBA_COMB LOFAR_CORE_HBA_SEP
#  LOFAR_REMOTE_LBA LOFAR_REMOTE_HBA
#  LOFAR_INTERNATIONAL_LBA
#  LOFAR_INTERNATIONAL_HBA_COMB LOFAR_INTERNATIONAL_HBA_COMB
#  LOFAR_NL_LBA LOFAR_NL_HBA_COMB LOFAR_NL_HBA_SEP
#  MERLIN6 MERLIN7
#  MWA32T MWA512
#  HSA
#  VLA_A VLA_B VLA_C VLA_D VLA_ANB VLA_BNC VLA_CND
#  VLBA
#
# Version: 20110921

import arrays
import pyrap.tables
from pyrap.tables import tablecreatescalarcoldesc as cldsc
from pyrap.tables import tablecreatearraycoldesc as clarrdsc
import sys
import os.path

# Use an ordered dictionary so table columns are in the expected order
class ord_dict(dict):
    def __init__(self):
        dict.__init__(self)
        self._order = []
    def __setitem__(self, key, value):
        self._order.append(key)
        dict.__setitem__(self, key, value)
    def keys(self):
        return self._order
    def __iter__(self):
        self._iter_n = 0
        return self
    def next(self):
        if self._iter_n == len(self._order):
            raise StopIteration
        self._iter_n += 1
        return self._order[self._iter_n - 1]

# From pyrap but using ordered dictionary
def tablecreatedesc(descs=[]):
    rec = ord_dict()
    for desc in descs:
        colname = desc['name']
        if rec.has_key(colname):
            raise ValueError('Column name ' + name + ' multiply used in table description')
        rec[colname] = desc['desc']
    return rec

def gen_table(ms_out, ant_file, user_antennas):
    # Read requested antennas in to a dictionary indexed by ant code
    ant_list = {}
    inf = open(ant_file)
    for line in inf:
        if line.startswith('#') or line.strip() == '':
            continue
        name, station, x, y, z, mount, diameter = line.split()
        x, y, z, diameter = [float(i) for i in x, y, z, diameter]
        if name in user_antennas:
            ant_list[name] = {'NAME': name, 'STATION': station,
                              'POSITION': [x, y, z], 'MOUNT': mount,
                              'DISH_DIAMETER': diameter}
    inf.close()
    
    for ant in user_antennas:
        if ant not in ant_list:
            print >> sys.stderr, 'Warning: Antenna "%s" not found' % ant
    
    # Define columns
    offset_desc = clarrdsc('OFFSET', value=float(), ndim=1, shape=[3])
    position_desc = clarrdsc('POSITION', value=float(), ndim=1, shape=[3])
    type_desc = cldsc('TYPE', value=str())
    dish_desc = cldsc('DISH_DIAMETER', value=float())
    flag_desc = cldsc('FLAG_ROW', value=bool())
    mount_desc = cldsc('MOUNT', value=str())
    name_desc = cldsc('NAME', value=str())
    station_desc = cldsc('STATION', value=str())
    
    desc = tablecreatedesc([offset_desc, position_desc, type_desc,
            dish_desc, flag_desc, mount_desc, name_desc, station_desc])
    
    # Create and populate our table
    table = pyrap.tables.table(ms_out, desc, nrow=len(ant_list), readonly=False)
    table.putcolkeywords('OFFSET', {'QuantumUnits': ['m', 'm', 'm'],
        'MEASINFO': {'Ref': 'ITRF', 'type': 'position'}})
    table.putcolkeywords('POSITION', {'QuantumUnits': ['m', 'm', 'm'],
        'MEASINFO': {'Ref': 'ITRF', 'type': 'position'}})
    table.putcolkeywords('DISH_DIAMETER', {'QuantumUnits': ['m']})
    table[:] = {'DISH_DIAMETER': 25.0, 'OFFSET': [0.0,0.0,0.0], 'TYPE': 'GROUND-BASED'}
    for i in range(len(ant_list)):
        table[i] = ant_list[user_antennas[i]]

def get_antfile():
    prog_dir = os.path.dirname(__file__)
    if os.path.exists('./ants.txt') or prog_dir == '':
      prog_dir = '.'
    return prog_dir + '/ants.txt'

def print_eg():
    print >> sys.stderr, "eg: \"['EB', 'JB', 'MC', 'NT', 'TR']\""
    print >> sys.stderr, "    \"EVN + VLBA + ['AR', 'Y']\""
    print >> sys.stderr, "    \"LOFAR_REMOTE_LBA + ['CS401LBA', 'CS501LBA']\""

if __name__ == '__main__':
    if len(sys.argv) != 3:
        print >> sys.stderr, 'Usage: %s <out_table> "<antennas>"' % sys.argv[0]
	print_eg()
        sys.exit()
    
    ms_out = sys.argv[1]
    ant_desc = sys.argv[2]
    try:
        user_antennas = arrays.array_ants(ant_desc)
    except ValueError:
        print >> sys.stderr, 'Antenna list not valid: "%s"' % ant_desc
	print_eg()
        sys.exit()
        
    ant_file = get_antfile()
    gen_table(ms_out, ant_file, user_antennas)
