#!/usr/bin/env python

# This program can change certin values in a Measurement Set.
# Can be used with makems to to generate non-lofar MSs.
#
# Stephen Bourke, JIVE
# Dec, 2009

from pyrap.tables import table
import sys
import numpy

import warnings
warnings.filterwarnings('ignore', category=DeprecationWarning)

stokes = {'I': 1, 'Q': 2, 'U': 3, 'V': 4,
          'RR': 5, 'RL': 6, 'LR': 7, 'LL': 8,
          'XX': 9, 'XY': 10, 'YX': 11, 'YY': 12}

def circ_feed(msname):
	"""Set the antenna feeds to L,R."""
	feed = table(msname + '/FEED', readonly=False)
	feed[:] = {'POLARIZATION_TYPE': ['L', 'R']}

def circ_pol(msname):
	"""Set the polarisation of the MS to circular."""
	pol = table(msname + '/POLARIZATION', readonly=False)
	if len(pol[0]['CORR_TYPE']) == 2:
		stk = [stokes['RR'], stokes['LL']]
	elif len(pol[0]['CORR_TYPE']) == 4:
		stk = [stokes['RR'], stokes['RL'], stokes['LR'], stokes['LL']]
	else:
		raise ValueError, 'CORR_TYPE != 2 or 4'
	pol[:] = {'CORR_TYPE': stk}

def circ(msname):
	"""Set the MS to circular polarisation. This does not actully
	alter the data."""
	circ_feed(msname)
	circ_pol(msname)

def field(msname, id, field_name):
	"""Set field names."""
	field = table(msname + '/FIELD', readonly=False)
	field[int(id) - 1] = {'NAME': field_name}

def tel(msname, tel_name):
	"""Set the telescope name."""
	obs = table(msname + '/OBSERVATION', readonly=False)
	obs[:] = {'TELESCOPE_NAME': tel_name}

def sched(msname, sched_name):
	"""Set the schedule type (probably not important, makems set it
	to LOFAR by default."""
	obs = table(msname + '/OBSERVATION', readonly=False)
	obs[:] = {'SCHEDULE_TYPE': sched_name}

def cols(msname):
	"""Add Columns not inserted by makmes"""
	ms = table(msname, readonly=False)
	channels, pols = ms.getcoldesc('DATA')['shape']
	model_desc = ms.getcoldesc('DATA')
	model_desc['comment'] = 'model data'
	model_desc['shape'] = model_desc['shape'][::-1] # workaround pyrap bug
	ms.addcols({'MODEL_DATA': model_desc})
	corr_desc = ms.getcoldesc('DATA')
	corr_desc['comment'] = 'corrected data'
	corr_desc['shape'] = corr_desc['shape'][::-1] # workaround pyrap bug
	ms.addcols({'CORRECTED_DATA': corr_desc})
	imagwt_desc = ms.getcoldesc('WEIGHT')
	imagwt_desc['comment'] = 'imaging weight'
	imagwt_desc['shape'] = numpy.array([channels])
	ms.addcols({'IMAGING_WEIGHT': imagwt_desc})

	# Reopen MS to flush data.
	ms = table(msname, readonly=False)
	data = ms[0]['DATA']
	data[:,:] = 1j # Same default as lwimager
	weights = numpy.ones(shape=(channels,), dtype='float32')
	ms[:] = {'MODEL_DATA': data}
	ms[:] = {'CORRECTED_DATA': data}
	ms[:] = {'IMAGING_WEIGHT': weights}
	ms.putcolkeywords('MODEL_DATA', {'CHANNEL_SELECTION': numpy.array([[0,channels]])})

if __name__ == '__main__':
	if len(sys.argv) < 2:
		print >> sys.stderr, 'Usage: %s <function> [args]' % sys.argv[0]
		print >> sys.stderr, 'See source code for functions & args'
		sys.exit()
	func = eval(sys.argv[1])
	args = sys.argv[2:]
	func(*args)
