#!/usr/bin/env python

# Program to calculate the RA and DEC of the zenith for the
# current time of the Latitude and Longtitude supplied on the
# command line. Alternatively, a Measurement Set Antenna table
# can be supplied and it will calculate the mean antenna
# position and use that as the Latitude, Longtitude.
#
# Stephen Bourke, JIVE.
# Oct, 2009.

from datetime import datetime
import math
import sys

import warnings
warnings.filterwarnings('ignore', category=DeprecationWarning)

def j2000days(datetm):
	"""Returns fractional days from J2000 to datetm"""
	j2000 = datetime(2000, 1, 1, 12, 0, 0)
	delta = datetm - j2000
	return delta.days + delta.seconds / float(24 *60 * 60)

def hours(datetm):
	"""Returns fractional hours in datetm"""
	return datetm.hour + datetm.minute / 60.0 + datetm.second / 3600.0

def lst(datetm, lon):
	"""Returns Local Siderial Time"""
	lt = 100.46 + 0.985647 * j2000days(datetm) + lon + 15 * hours(datetm)
	return lt % 360

def dms2deg(dms):
	"""Returns a fractional degree from dd mm ss"""
        deg = abs(dms[0]) + float(dms[1]) / 60 + float(dms[2]) / 3600
	if dms[0] < 0:
		deg *= -1
	return deg

def deg2dms(angle):
	"""Returns the dd mm ss tuple of angle"""
        minutes, degrees = math.modf(angle)
        minutes = abs(minutes * 60)
        seconds, minutes = math.modf(minutes)
        seconds = seconds * 60
        return degrees, minutes, seconds

def cart2sph(coord):
	"""Convert from cartesian to spherical coordinates."""
        x,y,z = coord
        r = math.sqrt(x**2 + y**2 + z**2)
        lat = -1 * (math.atan(math.sqrt(x**2 + y**2)/z) - (math.pi / 2))
        lon = math.atan2(y,x)
        if lat > math.pi / 2:
                lat -= math.pi
        return r, lon, lat

def lat_long_from_ms(msant):
	"""Return the Longtitude and Latitude of the mean antenna position."""
	import pyrap.tables
	ant_table = pyrap.tables.table(msant, ack=False)
	x0 = y0 = z0 = 0
	num_ants = len(ant_table)
	for row in ant_table:
		x, y, z = row['POSITION']
		x0 += x / num_ants
		y0 += y / num_ants
		z0 += z / num_ants
	r, lon, lat = cart2sph((x0, y0, z0))
	return [math.degrees(i) for i in (lon, lat)]

if __name__ == '__main__':
	if len(sys.argv) == 2:
		# Argument is the name of an MS Antenna Table
		longtitude, latitude = lat_long_from_ms(sys.argv[1])
	elif len(sys.argv) == 3:
		# Arguments are RA, DEC, eg 4.5 30.66
		latitude = float(sys.argv[1])
		longtitude = float(sys.argv[2])
	elif len(sys.argv) == 7:
		# Arguments are RA, DEC, eg 4 30 0 30 40 0
		latitude = dms2deg([float(i) for i in sys.argv[1:4]])
		longtitude = dms2deg([float(i) for i in sys.argv[4:7]])
	else:
		print >> sys.stderr, 'Usage: %s [AntTable | <lat> <lon>]' % sys.argv[0]
		sys.exit()
	obs_time = datetime.utcnow()
	ra_deg = lst(obs_time, longtitude)
	yy, mm, dd = obs_time.year, obs_time.month, obs_time.day
	h, m, s = obs_time.hour, obs_time.minute, obs_time.second
	print 'StartTime=%04d/%02d/%02d/%02d:%02d:%02d' % (yy, mm, dd, h, m, s)
	print 'RightAscension=%02d:%02d:%04.1f' % deg2dms(ra_deg / 15 % 24)
	print 'Declination=%02d.%02d.%04.1f' % deg2dms(latitude)
