#!/usr/bin/env python2.4

from datetime import *
from calendar import timegm
import ms
import re
import optparse

# General strategy
# ----------------
#
# There may be more than one GAIN group for a single antenna.  In
# particular, there may be seperate GAIN groups for RCP and LCP.
# Since we may have to combine these into a single row in the GAIN
# table, we must parse the entire file before we can write them into
# the MeasurementSet.
#
# There can probably be more TSYS and TANT groups too, but these can
# be written immediately since we don't have to combine them.
#
# Reporting errors and warnings
# -----------------------------
#
# An error message is issued on conditions that will cause data from
# the ANTAB file not to be written to the MeasurementSet.  Warning
# messages are generated for conditions that are irregular, but would
# not lead to loss of data.
#

# Regular expression to match a GAIN group.
#
# Group 1 is the list of additional keywords.
gain_re = re.compile(r"^GAIN\s+(.*)", re.I)

# Regular expressions to match required and optional keywords in a
# GAIN group.
#
# DPFU; Group 1 is the zenith point sensitivity.
dpfu_re = re.compile(r"DPFU\s*=\s*([-+\d.E,]+)", re.I)
# EQUAT|ALTAZ|ELEV|GCNRAO; Group 1 is the mount type.
type_re = re.compile(r"(EQUAT|ALTAZ|ELEV|GCNRAO)", re.I)
# TABLE; Tabulated values immediately follow the GAIN group.
table_re = re.compile(r"TABLE", re.I)
# POLY; Group 1 specifies the coefficients.
poly_re = re.compile(r"POLY\s*=\s*([-+\d.E,]+)", re.I)
# OFFSET; Group 1 specifies the offset.
offset_re = re.compile(r"OFFSET\s*=\s*([-+\d.E])", re.I)
# DEC; Group 1 specifies the declination.
dec_re = re.compile(r"DEC\s*=\s*([-+\d.E]+)", re.I)
# RCP|LCP; Group 1 is the polarization type.
polarization_re = re.compile(r"(RCP|LCP)", re.I)
# FREQ; Group 1 and 2 are the lower and upper bound of the frequency
# interval.  Apparantly frequencies can be negative.
freq_re = re.compile(r"FREQ\s*=\s*([-+\d.E]+),([-+\d.E]+)")
# A00|A10|A11E|A11O|A20|A21E|A21O|A22E|A22O|A30; Group 1 is the
# coefficient.
spherical_re = [
    re.compile(r"A00\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A10\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A11E\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A11O\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A20\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A21E\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A21O\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A22E\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A22O\s*=\s*([-+\d.E]+)", re.I),
    re.compile(r"A30\s*=\s*([-+\d.E]+)", re.I)]

# Dictionary translating the mount types.
type_dict = {'EQUAT': 'HA',
             'ALTAZ': 'ZA',
             'ELEV':  'EL' }

# Regular expression to match a TSYS group.
#
# Group 1 is the list of additional keywords.
tsys_re = re.compile(r"^TSYS\s+(\w+)\s+(.*)", re.I)

# Regular expressions to match optional keywords in a TSYS or TANT
# group.
#
# FT; Group 1 is the relevant factor.
ft_re = re.compile(r"FT\s*=\s*([-+\d.E]+)", re.I)
# TIMEOFF; Group 1 is the time offset.
timeoff_re = re.compile(r"TIMEOFF\s*=\s*([-+\d.E]+)", re.I)

# Regular expression to match a column description line.
#
# Group 1 is the column number, group 2 is the polarization (either R
# or L), group 3 is the frequency and group 4 is the bandwidth.
column_re = re.compile(r"^!Column\s*(\d+)\s*=\s*(R|L)\d+:.*,\s+"
                       r"(\d+\.\d+)\s*MHz,\s+"
                       r"BW\s*=\s*(\d+\.\d+)\s*MHz")
# Group 1 is the sideband.  Not included above since it might be absent.
sideband_re = re.compile(r"(USB|LSB)", re.I)

# Regular expression to match temperature value line.
#
# Group 1 is the day, group 2 is the hour, group 3 the minutes and
# group 4 the temperature values.
temperature_re = re.compile(r"^(\d+)\s+(\d\d):(\d\d\.\d\d)\s+(.*)")

# Regular expression to match group delimiters.
delimiter_re = re.compile(r"/\s*$")


# Antenna's for which we have gain curves.
gain_antenna_names = {}

# Antenna's for which we have system temberatures.
tsys_antenna_names = {}


#####################
# Helper functions. #
#####################

# Skip until the next delimiter.
#
def skip_block(input):

    while 1:
        # Read next line.
        line = input.readline()
        if not line:
            return

        delimiter_match = delimiter_re.search(line)
        if delimiter_match:
            return


# Cache for antenna IDs.
antenna_ids = {}
antenna_name = {}

def get_antenna_ids(ms, antenna):

    if not antenna in antenna_ids:
        antenna_ids[antenna] = ms.get_antenna_ids(antenna)
        for antenna_id in antenna_ids[antenna]:
            antenna_name[antenna_id] = antenna

        # Print an error message if we didn't find the antanna.
        if len(antenna_ids[antenna]) == 0:
            print 'Error: missing antenna %s' % antenna

        # Print a warning message if we have multiple matches.    
        if len(antenna_ids[antenna]) > 1:
            print 'Warning: multiple antennas %s found' % antenna

    return antenna_ids[antenna]


# Cache for spectral window IDs.
spwindow_ids = {}
spwindow_frequency = {}

def get_spwindow_ids(ms, frequency, tol = 1e3):

    if not frequency in spwindow_ids:
        spwindow_ids[frequency] = ms.get_spwindow_ids(frequency, tol)
        for spwindow_id in spwindow_ids[frequency]:
            spwindow_frequency[spwindow_id] = frequency

        # Print an error message if we didn't find the spectral window.
        if len(spwindow_ids[frequency]) == 0:
            mhz = frequency / 1e6
            print 'Error: missing spectral window for %f MHz frequency' % mhz

    return spwindow_ids[frequency]


# Cache for feed IDs.
feed_ids = {}

def get_feed_ids(ms, antenna_id, spwindow_id):

    if not (antenna_id, spwindow_id) in feed_ids:
        ids = ms.get_feed_ids(antenna_id, spwindow_id)
        feed_ids[(antenna_id, spwindow_id)] = ids

        mhz = spwindow_frequency[spwindow_id] / 1e6
        antenna = antenna_name[antenna_id]

        # Print an error message if we didn't find the feed.
        if len(feed_ids[(antenna_id, spwindow_id)]) == 0:
            print 'Error: missing feed for %f MHz frequency on antenna %s' \
                  % (mhz, antenna)

        # Print a warning message if we have multiple matches.
        if len(feed_ids[(antenna_id, spwindow_id)]) > 1:
            print 'Warning: multiple feeds found for %f MHz frequency' \
                  ' on antenna %s' % (mhz, antenna)

    return feed_ids[(antenna_id, spwindow_id)]


# Cache for polarization types.
pol_type = {}

def get_pol_type(ms, antenna_id, feed_id, spwindow_id):

    if not (antenna_id, feed_id, spwindow_id) in pol_type:
        type = ms.get_pol_type(antenna_id, feed_id, spwindow_id)
        pol_type[(antenna_id, feed_id, spwindow_id)] = type

    return pol_type[(antenna_id, feed_id, spwindow_id)]


#######################
# Gain curve support. #
#######################

# Dictionary of gain curves.  This will become a multi-level
# dictionary indexed by antenna name, frequency range, and attributes.
# Currently, the following attributes are used:
#
# 'type': 'TABLE' (tabulated) or 'POLY' (polynomial).
#
# 'subtype': Coordinate system ('ZA', 'EL', 'ZAAZ', 'ELAZ', 'HA',
#            'HADEC' or 'SPHER').
#
# 'gain': Nominal peak sensitivity (K/Jy).
#
# 'coords': List of tabulated coordinates (optional).
#
# 'decs': List of tabulated declinations (optional).
#
# 'data': Tabulated values or coefficients.
#
# The 'data' and 'gain' attributes are further indexed by polarization.
#
gain_curve = {}

def add_gain_curve(antenna, freq_range, polarization, dec,
                   gain, type, subtype, coords, data):

    if not antenna in gain_curve:
        gain_curve[antenna] = {}

    if not freq_range in gain_curve[antenna]:
        gain_curve[antenna][freq_range] = {}
        gain_curve[antenna][freq_range]['type'] = type
        gain_curve[antenna][freq_range]['subtype'] = subtype
        gain_curve[antenna][freq_range]['gain'] = {}
        if coords:
            gain_curve[antenna][freq_range]['coords'] = coords
        if dec:
            gain_curve[antenna][freq_range]['decs'] = []
        gain_curve[antenna][freq_range]['data'] = {}
    else:
        # Consistency checks.
        #
        if not gain_curve[antenna][freq_range]['type'] == type:
            print 'Error: Inconsistent gain curve specifications' \
                  ' for antenna %s' % antenna
            return
        if not gain_curve[antenna][freq_range]['subtype'] == subtype:
            print 'Error: Inconsistent gain curve specifications' \
                  ' for antenna %s' % antenna
            return

    # If a declination was specified, we need yet another level in the
    # gain_curve dictionary.  We also need to keep track of the
    # specified declinations.
    #
    if dec:
        if not dec in gain_curve[antenna][freq_range]['decs']:
            gain_curve[antenna][freq_range]['decs'].append(dec)
        if not polarization in gain_curve[antenna][freq_range]['gain']:
            gain_curve[antenna][freq_range]['data'][polarization] = {}
        gain_curve[antenna][freq_range]['data'][polarization][dec] = data
    else:
        gain_curve[antenna][freq_range]['data'][polarization] = data
    gain_curve[antenna][freq_range]['gain'][polarization] = gain


# Parse tabulated gain values.
#
def parse_gain_table(input):

    coords = []
    table = []

    while 1:
        # Read next line.
        line = input.readline()
        if not line:
            break

        # Strip out comments.
        index = line.find('!')
        if index != -1:
            line = line[:index]
        if not line:
            continue

        row = line.split()
        if (len(row) == 2):
            coords.append(float(row[0]))
            table.append(float(row[1]))

        delimiter_match = delimiter_re.search(line)
        if delimiter_match:
            break

    return (coords, table)


# Parse GAIN group, including the list of values which follow it.
#
def parse_gain(input, ms, keywords, options):

    antenna = None
    dpfu = None
    type = None
    table = False
    poly = None
    offset = 0                          # Default value.
    dec = None
    polarizations = ['R', 'L']          # Default value.
    freq = None
    data = 10 * [float('nan')]

    while keywords:
        antenna_match = antenna_re.search(keywords)
        if antenna_match:
            antenna = antenna_match.group(1)

        dpfu_match = dpfu_re.search(keywords)
        if dpfu_match:
            dpfu = map(float, dpfu_match.group(1).split(','))

        type_match = type_re.search(keywords)
        if type_match:
            type = type_match.group(1)

        table_match = table_re.search(keywords)
        if table_match:
            table = True;

        poly_match = poly_re.search(keywords)
        if poly_match:
            poly = map(float, poly_match.group(1).split(','))

        offset_match = offset_re.search(keywords)
        if offset_match:
            offset = float(offset_match.group(1))

        dec_match = dec_re.search(keywords)
        if dec_match:
            dec = float(dec_match.group(1))

        polarization_match = polarization_re.search(keywords)
        if polarization_match:
            polarizations = [polarization_match.group(1)[0]]

        freq_match = freq_re.search(keywords)
        if freq_match:
            low = float(freq_match.group(1)) * 1e6 # MHz
            high = float(freq_match.group(2)) * 1e6 # MHz
            freq = (low, high)

        for i in range(len(spherical_re)):
            spherical_match = spherical_re[i].search(keywords)
            if spherical_match:
                data[i] = float(spherical_match.group(1))

        delimiter_match = delimiter_re.search(keywords)
        if delimiter_match:
            # Validate the gain curve description.
            #
            if not antenna:
                print 'Error: missing or unrecognized antenna name ' \
                      'in GAIN group'
                return

            if options.verbose:
                print 'Processing gain curve for %s' % antenna

            if not type:
                print 'Error: missing gain curve type' \
                      ' for antenna %s' % antenna
                return

            if not dpfu:
                print 'Error: missing zenith point source sensitivity' \
                      ' for antenna %s' % antenna
                return

            if len(dpfu) > 2:
                print 'Error: source density for more than two polarizations' \
                      ' on antenna %s' % antenna
                return

            if not offset == 0:
                print 'Error: unsupported offset %d for antenna %s' \
                      % (offset, antenna)
                return

            if dec != None and type != 'EQUAT':
                print 'Error: specification of declination for %s' \
                      ' gain curve for antenna %s' % (type, antenna)
                return

            if not freq:
                print 'Warning: missing frequency range' \
                      ' for antenna %s' % antenna
                # Synthesize a frequency range; 0 Hz - 1 THz should
                # cover all radio bands.
                freq = (0, 1e12)

            # Ready to process this gain curve.
            #
            if len(dpfu) == 1:
                dpfu.append(dpfu[0])

            if type == 'EQUAT' and dec == None:
                dec = float('nan')

            if poly:
                subtype = type_dict[type]
                type = 'POLY'
                coords = None
                data = poly
            elif table:
                subtype = type_dict[type]
                type = 'TABLE'
                (coords, data) = parse_gain_table(input)
            elif type == 'GCNRAO':
                type = 'POLY'
                subtype = 'SPHER'
                coords = None
            else:
                print 'Error: unsupported gain curve type %s' \
                      ' for antenna %s' % (type, antenna)
                return

            i = 0
            for polarization in polarizations:
                add_gain_curve(antenna, freq, polarization, dec,
                               dpfu[i], type, subtype, coords, data)
                i += 1

            return

        keywords = input.readline()


# Transform a list of vectors (one per recepetor) into a proper matrix
# with one column per receptor.
#
# XXX The receptor (polarization) should probably be the first array
# index instead of the last.  That would make life much easier since
# it'd remove the need to do this transformation (although we should
# still make sure the vectors are of equal length).
#
def transform_vector_data(data):

    num_receptors = len(data)
    num_elements = 0
    for elements in data:
        num_elements = max(num_elements, len(elements))

    # Make sure all vectors are of equal length.
    #
    for elements in data:
        while len(elements) < num_elements:
            elements.append(float('nan'))

    return [[element[i] for element in data] for i in range(num_elements)]


# Transform a list of lists of vectors (one per recepetor and
# declination) into a proper cube.
#
# XXX The receptor (polarization) should probably be the first array
# index instead of the last.  That would make life much easier since
# it'd remove the need to do this transformation (although we should
# still make sure the vectors are of equal length).
#
def transform_matrix_data(data, decs):

    num_receptors = len(data)
    num_decs = len(decs)
    num_elements = 0
    for receptor in data:
        for dec in receptor:
            num_elements = max(num_elements, len(receptor[dec]))

    for receptor in data:
        for dec in decs:
            if not receptor[dec]:
                receptor[dec] = []
        for dec in receptor:
            while len(receptor[dec]) < num_elements:
                receptor[dec].append(float('nan'))

    return [[[element[dec][i] for element in data] for dec in decs] \
            for i in range(num_elements)]


# Write a single gain curve to the GAINCURVE table in the
# MeasurementSet.
#
def write_gain_curve(ms, antenna_id, feed_id, spwindow_id, gain_curve):

    # Make sure the data matches the polarizations for the antenna
    # feed, sorting them in the right order, inserting NaN's for
    # missing data.
    #
    sorted_gains = []
    sorted_data = []
    pol_type = get_pol_type(ms, antenna_id, feed_id, spwindow_id)
    for polarization in pol_type:
        if polarization in gain_curve['gain']:
            sorted_gains.append(gain_curve['gain'][polarization])
            sorted_data.append(gain_curve['data'][polarization])
        else:
            sorted_gains.append(float('nan'))
            sorted_data.append([float('nan')])

    # Build a proper gain curve record from the data.
    curve = {'type': gain_curve['type']}
    if curve['type'] == 'POLY':
        curve['polytype'] = gain_curve['subtype']
        if curve['polytype'] == 'ZA' or curve['polytype'] == 'EL' \
               or curve['polytype'] == 'SPHER':
            curve['coeffs'] = transform_vector_data(sorted_data)
        elif curve['polytype'] == 'HA':
            curve['decs'] = gain_curve['decs']
            curve['coeffs'] = transform_matrix_data(sorted_data, curve['decs'])
        else:
            print 'Error: unsupported polynomial %s' % curve['polytype']
            return
    else:
        curve['tabletype'] = gain_curve['subtype']
        if curve['tabletype'] == 'ZA':
            curve['za'] = gain_curve['coords']
            curve['table'] = transform_vector_data(sorted_data)
        elif curve['tabletype'] == 'EL':
            curve['el'] = gain_curve['coords']
            curve['table'] = transform_vector_data(sorted_data)
        elif curve['tabletype'] == 'HA':
            curve['tabletype'] = 'HADEC'
            curve['ha'] = gain_curve['coords']
            curve['dec'] = gain_curve['decs']
            curve['table'] = transform_matrix_data(sorted_data, curve['dec'])
        else:
            print 'Error: unsupported table %s' % curve['tabletype']
            return

    ms.add_gain_entry(antenna_id, feed_id, spwindow_id, sorted_gains, curve)


# Write gain curves to the GAINCURVE table in the MeasurementSet.
#
def write_gain_curves(ms, options):

    for antenna in gain_curve:

        antenna_ids = get_antenna_ids(ms, antenna)
        if not antenna_ids:
            continue

        if options.verbose:
            print 'Writing gain curve for %s' % antenna

        for freq_range in gain_curve[antenna]:

            frequency = (freq_range[0] + freq_range[1]) / 2
            bandwidth = freq_range[1] - freq_range[0]
            spwindow_ids = get_spwindow_ids(ms, frequency, bandwidth / 2)
            gains = gain_curve[antenna][freq_range]['gain']
            data = gain_curve[antenna][freq_range]['data']

            for antenna_id in antenna_ids:
                for spwindow_id in spwindow_ids:

                    # Look up the feed.  We allow for multiple feed
                    # IDs, but a warning will be printed.
                    #
                    feed_ids = get_feed_ids(ms, antenna_id, spwindow_id)
                    for feed_id in feed_ids:
                        write_gain_curve(ms, antenna_id, feed_id, spwindow_id,
                                         gain_curve[antenna][freq_range])

                        # Record that we have data for ANTENNA.
                        gain_antenna_names[antenna.upper()] = True


######################
# Tsys/Tant support. #
######################

# Add Tsys/Tant to the SYSCAL table in the MeasurementSet.
#
def add_temperature(ms, antenna, frequency, secs,
                    polarizations, temperatures):

    # Lookup antenna and spectral window.  We allow for multiple
    # antenna and spectral window IDs, but a warning will be printed
    # when multiple antenna IDs are found.  Multiple spectral window
    # IDs are OK, so don't print anything.
    #
    antenna_ids = get_antenna_ids(ms, antenna)
    spwindow_ids = get_spwindow_ids(ms, frequency)
    for antenna_id in antenna_ids:
        for spwindow_id in spwindow_ids:

            # Look up the feed.  We allow for multiple feed IDs, but a
            # warning will be printed.
            #
            feed_ids = get_feed_ids(ms, antenna_id, spwindow_id)
            for feed_id in feed_ids:

                # Make sure the temperature data matches the
                # polarizations for the antenna feed, sorting them in
                # the right order, inserting NaN's for missing
                # temperatures.
                #
                sorted_temperatures = []
                pol_type = get_pol_type(ms, antenna_id, feed_id, spwindow_id)
                for polarization in pol_type:
                    if polarization in polarizations:
                        index = polarizations.index(polarization)
                        sorted_temperatures.append(temperatures[index])
                    else:
                        sorted_temperatures.append(float('nan'))

                ms.add_syscal_entry(antenna_id, feed_id, spwindow_id,
                                    secs, sorted_temperatures)


# Parse list of Tsys/Tant values.
#
def parse_temperatures(input, ms, antenna, ft, timeoff, options):

    # Set to true if we've told the user we're writing the system
    # temperatures.
    msg = False

    # Set to false if we've complained about missing column
    # descriptions already.
    complain = True

    # Clear dictionary
    columns = {}

    while 1:
        # Read next line.
        line = input.readline()
        if not line:
            return

        # Build a dictionary of tuples (polarization, frequency,
        # bandwidth, sideband) with the column number as a key.
        #
        # Determination of the proper sideband is somewhat tricky.
        # The information is only present in recent EVN Calibrarion
        # Information Files.  If the information is absent, we employ
        # the following heuristics:
        #
        # If we encounter two identical column descriptions for the
        # same column, we assume the first is LSB and the second is
        # USB.  Otherwise, we assume DSB and hope that only one
        # spectral window on a single antenna feed matches the given
        # frequency.
        #
        column_match = column_re.match(line)
        if column_match:
            index = int(column_match.group(1))
            polarization = column_match.group(2)
            frequency = float(column_match.group(3)) * 1e6 # MHz
            bandwidth = float(column_match.group(4)) * 1e6 # MHz
            sideband_match = sideband_re.search(line)
            if sideband_match:
                sideband = sideband_match.group(1)
            else:
                sideband = 'DSB'
            entry = (polarization, frequency, bandwidth, sideband)

            if not index in columns:
                columns[index] = []
            if not entry in columns[index]:
                columns[index].append(entry)
            elif sideband == 'DSB':
                del columns[index][columns[index].index(entry)]
                entry = (polarization, frequency, bandwidth, 'LSB')
                columns[index].append(entry)
                entry = (polarization, frequency, bandwidth, 'USB')
                columns[index].append(entry)
            else:
                print 'Error: duplicate column description encountered' \
                      ' for antenna %s' % antenna

            continue

        
        # Strip out comments.
        index = line.find('!')
        if index != -1:
            line = line[:index]
        if not line:
            continue

        # Process temperature values, but only of we have a
        # description for the columns.
        #
        temperature_match = temperature_re.match(line)
        if temperature_match:

            if not len(columns) > 0:
                if complain:
                    print 'Error: missing column descriptions' \
                          ' for antenna %s' % antenna
                    complain = False
                continue
                
            # Build timestamp.
            #
            day = int(temperature_match.group(1))
            day = date(options.year, 1, 1).toordinal() + day - 1
            d = date.fromordinal(day)
            hour = int(temperature_match.group(2))
            t = time(hour)
            minute = float(temperature_match.group(3))
            d = datetime.combine(d, t) + timedelta(minutes = minute)

            # Convert timestamp to number of seconds since 00:00:00
            # UTC, January 1, 1970.
            #
            secs = timegm(d.utctimetuple()) + timeoff
            # Use microsecond precision, otherwise two entries might
            # end up with identical timestamps.
            secs += d.microsecond * 1e-6

            # Create a list of the temperatures.
            #
            temperatures = temperature_match.group(4).split()

            # Build a dictionary of tuples (polarization,
            # temperature), with the center frequency as a key.
            #
            data = {}                   # Clear dictionary.
            index = 1
            for temperature in temperatures:
                for description in columns[index]:
                    polarization = description[0]
                    ref_freq = description[1]
                    bandwidth = description[2]
                    sideband = description[3]

                    # Calculate the center frequency.
                    #
                    if sideband == 'USB':
                        center_freq = ref_freq + bandwidth / 2
                    elif sideband == 'LSB':
                        center_freq = ref_freq - bandwidth / 2
                    else:
                        center_freq = 0
                    
                    if not center_freq in data:
                        data[center_freq] = []
                    temperature = float(temperature) * ft
                    data[center_freq].append((polarization, temperature))
                index = index + 1

            for frequency in data:
                polarizations = [pol for (pol, temp) in data[frequency]]
                temperatures = [temp for (pol, temp) in data[frequency]]

                if options.tsys:
                    if options.verbose and not msg:
                        print 'Writing system temperatures for %s' % antenna
                        msg = True

                    add_temperature(ms, antenna, frequency, secs,
                                    polarizations, temperatures)

                    # Record that we have data for ANTENNA.
                    tsys_antenna_names[antenna.upper()] = True

        delimiter_match = delimiter_re.search(line)
        if delimiter_match:
            return


# Parse TSYS group, including the list of values which follow it.
#
def parse_tsys(input, ms, keywords, options):

    ft = 1.0                            # Default value.
    timeoff = 0.0                       # Default value.

    while keywords:
        antenna_match = antenna_re.search(keywords)
        if antenna_match:
            antenna = antenna_match.group(1)

        ft_match = ft_re.search(keywords)
        if ft_match:
            ft = float(ft_match.group(1))

        timeoff_match = timeoff_re.search(keywords)
        if timeoff_match:
            timeoff = float(timeoff_match.group(1))

        delimiter_match = delimiter_re.search(keywords)
        if delimiter_match:
            if not antenna:
                print 'Error: missing or unrecognized antenna name ' \
                      'in TSYS group'
                skip_block(input)
                return

            if options.verbose:
                print 'Processing system temperatures for %s' % antenna
                
            parse_temperatures(input, ms, antenna, ft, timeoff, options)
            return

        keywords = input.readline()


########################
# Program entry point. #
########################

usage = 'usage: %prog [options] antab ms'
parser = optparse.OptionParser(usage=usage, version='%prog 0.91')
parser.add_option('-c', '--clear', action="store_true", dest='clear',
                  default=False, help='clear table before adding entries')
parser.add_option('-g', '--gain-curve', action="store_true", dest='gain',
                  default=False, help='add gain curves')
parser.add_option('-s', '--system-temperature', action="store_true",
                  dest='tsys', default=False, help='add system temperatures')
parser.add_option('-v', '--verbose', action="store_true", dest='verbose',
                  default=True, help='be verbose')
parser.add_option('-q', '--quiet', action="store_true", dest='verbose',
                  help='be quiet')
parser.add_option('-y', '--year', type='int', dest='year',
                  help='assume measurements are for YEAR', metavar='YEAR')
(options, args) = parser.parse_args()
if not options.gain and not options.tsys:
    options.gain = options.tsys = True
if len(args) != 2:
    parser.error("incorrect number of arguments")

ms = ms.MeasurementSet(args[1])
if options.clear:
    ms.clear_gain()
    ms.clear_syscal()

# Get year from MeasurementSet.
if not options.year:
    observation = datetime.utcfromtimestamp(ms.get_observation_time())
    options.year = observation.year

# Build regular expression to match antenna names.
antenna_names = ms.get_antenna_names();
antenna_re = '('
for antenna in antenna_names:
    antenna_re = antenna_re + antenna.upper() + '|'
antenna_re = antenna_re[:-1] + ')'
antenna_re = re.compile(antenna_re, re.I)

input = open(args[0], 'r')
while 1:
    # Read next line.
    line = input.readline()
    if not line:
        break

    # Strip out comments.
    index = line.find('!')
    if index != -1:
        line = line[:index]
    if not line:
        continue

    gain_match = gain_re.match(line)
    if gain_match:
        parse_gain(input, ms, gain_match.group(1), options)
        continue

    tsys_match = tsys_re.match(line)
    if tsys_match:
        parse_tsys(input, ms, tsys_match.group(1), options)
        continue

input.close()

if options.gain:
    write_gain_curves(ms, options)

# Check whether we have data for all antennas in the MeasurementSet.
for antenna in antenna_names:
    if not gain_antenna_names.get(antenna.upper(), False):
        print 'Warning: no gain curves for antenna %s' \
              % antenna.upper()
    if not tsys_antenna_names.get(antenna.upper(), False):
        print 'Warning: no system temperatures for antenna %s' \
              % antenna.upper()
