#!/usr/bin/env python2.4

from datetime import *
from calendar import timegm
import ms
import re
import optparse

# This code only accepts a subset of the UVFLG command set.  It only
# accepts the new format, i.e. with the ANT_NAME instead of the
# ANTENNAS adverb.  Actually, that, and the OPCODE, REASON and
# TIMERANG adverbs, are the only supported commands.

# Regular expression to match the DTIMRANG parameter.
#
# Regular expression group 1 is the parameter's value.
dtimrang_re = re.compile(r"DTIMRANG[\s=]+(\d+)", re.I)

# Regular expression to match the TIMEOFF parameter.
#
# Regular expression group 1 is the parameter's value.
timeoff_re = re.compile(r"TIMEOFF[\s=]+(\d+)", re.I)

# Regular expression to match the OPCODE adverb.
#
# Regular expression group 1 is the argument.
opcode_re = re.compile(r"OPCODE[\s=]+'([^']+)", re.I)

# Regular expression to match the ANT_NAME adverb.
#
# Regular expression group 1 is the antenna name.
antenna_re = re.compile(r"ANT_NAME[\s=]+'([^']+)", re.I)

# Regular expression to match the TIMERANG adverb.
#
# Regular expression group 1-4 give the start time, group 5-8 the end
# time of the interval covered.
timerang_re = re.compile(r"TIMERANG[\s=]+(\d+),(\d+),(\d+),(\d+),\s+"
                         r"(\d+),(\d+),(\d+),(\d+)", re.I)

# Regular expression to match the REASON adverb.
#
# Regular expression group 1 gives the reason.
reason_re = re.compile(r"REASON[\s=]+'([^']+)", re.I)

# Regular expression to match entry delimiter.
delimiter_re = re.compile(r"/\s*$")


# Cache for antenna IDs.
antenna_ids = {}
antenna_name = {}

def get_antenna_ids(ms, antenna):

    if not antenna in antenna_ids:
        antenna_ids[antenna] = ms.get_antenna_ids(antenna)
        for antenna_id in antenna_ids[antenna]:
            antenna_name[antenna_id] = antenna

        # Print an error message if we didn't find the antanna.
        if len(antenna_ids[antenna]) == 0:
            print 'Error: missing antenna %s' % antenna

        # Print a warning message if we have multiple matches.    
        if len(antenna_ids[antenna]) > 1:
            print 'Warning: multiple antennas %s found' % antenna

    return antenna_ids[antenna]


########################
# Program entry point. #
########################

usage = 'usage: %prog [options] uvflg ms'
parser = optparse.OptionParser(usage=usage, version='%prog 0.90')
parser.add_option('-c', '--clear', action="store_true", dest='clear',
                  default=False, help='clear table before adding entries')
parser.add_option('-y', '--year', type='int', dest='year',
                  help='assume measurements are for YEAR', metavar='YEAR')
(options, args) = parser.parse_args()
if len(args) != 2:
    parser.error("incorrect number of arguments")

ms = ms.MeasurementSet(args[1])
if options.clear:
    ms.clear_flagcmd()

# Get year from MeasurementSet.
if not options.year:
    observation = datetime.utcfromtimestamp(ms.get_observation_time())
    options.year = observation.year

input = open(args[0], 'r')
while 1:
    # Read next line.
    line = input.readline()
    if not line:
        break

    # Strip out comments.
    index = line.find('!')
    if index != -1:
        line = line[:index]
    if not line:
        continue

    dtimrang_match = dtimrang_re.search(line)
    if dtimrang_match:
        dtimrang = int(dtimrang_match.group(1))

    timeoff_match = timeoff_re.search(line)
    if timeoff_match:
        timeoff = int(timeoff_match.group(1))

    opcode_match = opcode_re.search(line)
    if opcode_match:
        opcode = opcode_match.group(1)

    antenna_match = antenna_re.search(line)
    if antenna_match:
        antenna = antenna_match.group(1)

    timerang_match = timerang_re.search(line)
    if timerang_match:
        # Initialize interval.
        interval = timedelta()

        # Build start time.
        #
        day = int(timerang_match.group(1))
        day = date(options.year, 1, 1).toordinal() + day - 1
        d = date.fromordinal(day)
        hour = int(timerang_match.group(2))
        minute = int(timerang_match.group(3))
        second = int(timerang_match.group(4))
        t = time(hour, minute, second)
        d1 = datetime.combine(d, t)

        # Build end time.
        #
        day = int(timerang_match.group(5))
        d = date.fromordinal(date(options.year, 1, 1).toordinal() + day - 1)
        hour = int(timerang_match.group(6))
        minute = int(timerang_match.group(7))
        second = int(timerang_match.group(8))
        # Deal with the fact that the input file may specify 60.
        if second == 60:
            second = 59
            interval = timedelta(seconds = 1)
        t = time(hour, minute, second)
        d2 = datetime.combine(d, t)

        # Calculate time interval and midpoint.
        #
        interval = d2 - d1 + interval
        midpoint = d1 + interval // 2

        # Convert midpoint and interval to number of seconds since
        # 00:00:00 UTC, January 1, 1970.
        #
        midpoint = timegm(midpoint.utctimetuple()) + midpoint.microsecond / 1e6
        midpoint = midpoint + timeoff
        interval = interval.seconds + interval.microseconds / 1e6
        interval = interval + 2 * dtimrang

    reason_match = reason_re.search(line)
    if reason_match:
        reason = reason_match.group(1)

    delimiter_match = delimiter_re.search(line)
    if delimiter_match:
        # Lookup antenna.  We allow for multiple antenna IDs, but
        # we'll issue a warning in that case.
        #
        ids = get_antenna_ids(ms, antenna)
        for antenna_id in ids:
            command = 'ANTENNA1 == %d || ANTENNA2 == %d' \
                      % (antenna_id, antenna_id)
            ms.add_flagcmd_entry(midpoint, interval, 'FLAG', reason, command)

        # Reset variables.
        #
        antenna = None
        midpoint = None
        reason = None

input.close()
