from pyrap_tables import table
from pyrap_quanta import quantity
from pyrap_tables import tablecreatescalarcoldesc as cldsc
from pyrap_tables import tablecreatearraycoldesc as clarrdsc
import numpy

import warnings
# Pyrap currently causes lots of Deprecation warnings.
warnings.filterwarnings('ignore', category=DeprecationWarning)

mod_jul_day = quantity(40587.0, 'd') # Modified Julien day

def set_unit(q, unit):
    """Set the unit type of a pyrap quantity."""
    q.convert(quantity(1.0, unit))

def get_scalar_col_unit(table, col_name):
    """Get the quantity unit of a scalar column."""
    unit = table.getcolkeyword(col_name, 'QuantumUnits')
    assert len(unit) == 1, 'Unexpected (non-scalar) data format'
    return unit[0]

class ord_dict(dict):
    """Minimal implementation of a dict that preserves order
    when iterating and when keys() is called."""
    def __init__(self):
        dict.__init__(self)
        self._order = []
    def __setitem__(self, key, value):
        self._order.append(key)
        dict.__setitem__(self, key, value)
    def keys(self):
        return self._order
    def __iter__(self):
        self._iter_n = 0
        return self
    def next(self):
        if self._iter_n == len(self._order):
            raise StopIteration
        self._iter_n += 1
        return self._order[self._iter_n - 1]

# From pyrap but using ordered dictionary to ensure that columns
# remain in the order they were provided in.
def tablecreatedesc(descs=[]):
    rec = ord_dict()
    for desc in descs:
        colname = desc['name']
        if rec.has_key(colname):
            raise ValueError('Column name ' + name + ' multiply used in table description')
        rec[colname] = desc['desc']
    return rec

class MeasurementSet(object):
    def __init__(self, msname, create_subtables=True):
        self.msname = msname
        self.create_subtables = create_subtables
        self.flagcmd_table = None
        self.gain_table = None
        self.syscal_table = None
        
    def has_table(self, table_name):
        """Returns True if the table exists and we can open it, otherwise False."""
        full_table_name = self.msname + '/' + table_name
        try:
            table(full_table_name, ack=False)
        except:
            return False
        else:
            return True

    def flagCmd(self):
        """Return the FLAG_CMD table object, creating if necessary and allowed."""
        if self.flagcmd_table != None:
            # Table is already open
            pass
        elif self.has_table('FLAG_CMD'):
            self.flagcmd_table = table(self.msname + '/FLAG_CMD', ack=False, readonly=False)
        elif self.create_subtables:
            self.initFlagCmd()
        else:
            raise ValueError, 'No FLAG_CMD Table present'
        return self.flagcmd_table

    def gainCurve(self):
        """Return the JIVE_GAINCURVE table object, creating if necessary and allowed."""
        if self.gain_table != None:
            # Table is already open
            pass
        elif self.has_table('JIVE_GAINCURVE'):
            self.gain_table = table(self.msname + '/JIVE_GAINCURVE', ack=False, readonly=False)
        elif self.create_subtables:
            self.initGainCurve()
        else:
            raise ValueError, 'No JIVE_GAINCURVE Table present'
        return self.gain_table

    def sysCal(self):
        """Return the SYSCAL table object, creating if necessary and allowed."""
        if self.syscal_table != None:
            # Table is already open
            pass
        elif self.has_table('SYSCAL'):
            self.syscal_table = table(self.msname + '/SYSCAL', ack=False, readonly=False)
        elif self.create_subtables:
            self.initSysCal()
        else:
            raise ValueError, 'No SYSCAL Table present'
        return self.syscal_table

    def initFlagCmd(self):
        """Create a FLAG_CMD table."""
        time_desc = cldsc('TIME', value=float())
        interval_desc = cldsc('INTERVAL', value=float())
        type_desc = cldsc('TYPE', value=str())
        reason_desc = cldsc('REASON', value=str())
        level_desc = cldsc('LEVEL', value=int())
        severity_desc = cldsc('SEVERITY', value=int())
        applied_desc = cldsc('APPLIED', value=bool())
        command_desc = cldsc('COMMAND', value=str())
        
        desc = tablecreatedesc([time_desc, interval_desc, type_desc,
            reason_desc, level_desc, severity_desc, applied_desc, command_desc])
        
        flagcmd_table = table(self.msname + '/FLAG_CMD', desc, nrow=0, readonly=False, ack=False)
        flagcmd_table.putcolkeywords('TIME', {'QuantumUnits': ['s'],
            'MEASINFO': {'type': 'epoch', 'Ref': 'UTC'}})
        flagcmd_table.putcolkeywords('INTERVAL', {'QuantumUnits': ['s']})
        #XXX Work around pyrap bug to add sub-table keyword
        flagcmd_path = flagcmd_table.name()
        del flagcmd_table
        self._add_subtable_keyword(flagcmd_path)
        self.flagcmd_table = table(self.msname + '/FLAG_CMD', readonly=False, ack=False)

    def initGainCurve(self):
        """Create a JIVE_GAINCURVE table."""
        receptors = self.get_num_receptors()
        ant_id_desc = cldsc('ANTENNA_ID', value=int())
        feed_id_desc = cldsc('FEED_ID', value=int())
        spwin_id_desc = cldsc('SPECTRAL_WINDOW_ID', value=int())
        gain_desc = clarrdsc('GAIN', value=float(), ndim=1, shape=[receptors])
        curve_desc = cldsc('CURVE', value=dict())
    
        desc = tablecreatedesc([ant_id_desc, feed_id_desc,
            spwin_id_desc, gain_desc, curve_desc])
    
        gain_table = table(self.msname + '/JIVE_GAINCURVE', desc, nrow=0, readonly=False, ack=False)
        #XXX Work around pyrap bug to add sub-table keyword
        gain_path = gain_table.name()
        del gain_table
        self._add_subtable_keyword(gain_path)
        self.gain_table = table(self.msname + '/JIVE_GAINCURVE', readonly=False, ack=False)
    
    def initSysCal(self):
        """Create a SYSCAL table with TSYS columns."""
        ant_id_desc = cldsc('ANTENNA_ID', value=int())
        feed_id_desc = cldsc('FEED_ID', value=int())
        spwin_id_desc = cldsc('SPECTRAL_WINDOW_ID', value=int())
        time_desc = cldsc('TIME', value=float())
        interval_desc = cldsc('INTERVAL', value=float())
        tsys_desc = clarrdsc('TSYS', value=float(), ndim=1, shape=[self.get_num_receptors()])
        tsys_flag_desc = cldsc('TSYS_FLAG', value=bool())
        
        desc = tablecreatedesc([ant_id_desc, feed_id_desc, spwin_id_desc,
            time_desc, interval_desc, tsys_desc, tsys_flag_desc])
        
        syscal_table = table(self.msname + '/SYSCAL', desc, nrow=0, readonly=False, ack=False)
        syscal_table.putcolkeywords('TIME', {'QuantumUnits': ['s'],
            'MEASINFO': {'type': 'epoch', 'Ref': 'UTC'}})
        syscal_table.putcolkeywords('INTERVAL', {'QuantumUnits': ['s']})
        #XXX Work around pyrap bug to add sub-table keyword
        syscal_path = syscal_table.name()
        del syscal_table
        self._add_subtable_keyword(syscal_path)
        self.syscal_table = table(self.msname + '/SYSCAL', readonly=False, ack=False)

    def _add_subtable_keyword(self, subtable_path):
        """Add a subtable keyword to the main table."""
	subtable_name = subtable_path.split('/')[-1]
        table_str = 'Table: ' + subtable_path
        main_table = table(self.msname, readonly=False, ack=False)
        #XXX Work around pyrap bug in table.putkeyword()
        main_table._putkeyword('', subtable_name, -1, False, table_str)

    def clear_flagcmd(self):
        """Delete all rows in the FLAG_CMD table."""
        flagcmd_table = self.flagCmd()
        flagcmd_table.removerows(range(len(flagcmd_table)))

    def clear_gain(self):
        """Delete all rows in the GAIN table."""
        gain_table = self.gainCurve()
        gain_table.removerows(range(len(gain_table)))

    def clear_syscal(self):
        """Delete all rows in the SYSCAL table."""
        syscal_table = self.sysCal()
        syscal_table.removerows(range(len(syscal_table)))

    def get_num_receptors(self, quiet=False):
        feed_table = table(self.msname + '/FEED', ack=False)
        receptors = list(set(feed_table.getcol('NUM_RECEPTORS')))
        if len(receptors) > 1 and not quiet:
            warnings.warn('Not all feeds have the same number of receptors, using %d' % recpetors)
        return int(max(receptors))
    
    def get_antenna_names(self):
        ant_table = table(self.msname + '/ANTENNA', ack=False)
        return ant_table.getcol('NAME')

    def get_antenna_ids(self, antenna):
        ant_table = table(self.msname + '/ANTENNA', ack=False)
        ids = []
        for id in range(len(ant_table)):
            if ant_table[id]['NAME'].upper() == antenna.upper():
                ids.append(id)
        return ids

    def get_spwindow_ids(self, frequency, tol):
        spwin_table = table(self.msname + '/SPECTRAL_WINDOW', ack=False)
        ids = []
        for id in range(len(spwin_table)):
             # CHAN_FREQ can be in accending or decending order
             f1 = spwin_table[id]['CHAN_FREQ'][0]
             f2 = spwin_table[id]['CHAN_FREQ'][-1]
             low = min(f1, f2) - tol
             high = max(f1, f2) + tol
             if low <= frequency <= high:
                 ids.append(id)
        return ids

    def get_feed_ids(self, antenna_id, spwindow_id):
        feed_table = table(self.msname + '/FEED', ack=False)
        ids = []
        # if no matches are found for spwindow_id then do for -1
        for spw_id in (spwindow_id, -1):
            for row in feed_table:
                if row['ANTENNA_ID'] == antenna_id and row['SPECTRAL_WINDOW_ID'] == spw_id:
                    ids.append(row['FEED_ID'])
            if spw_id == -1 or len(ids) != 0:
                break
        return ids

    def get_pol_type(self, antenna_id, feed_id, spwindow_id):
        """Return the polarization type. Eg: ['L', 'R']"""
        feed_table = table(self.msname + '/FEED', ack=False)
        pol_type = []
        # if no matches are found for spwindow_id then do for -1
        for spw_id in (spwindow_id, -1):
            for row in feed_table:
                if row['ANTENNA_ID'] == antenna_id \
                   and row['FEED_ID'] == feed_id \
                   and row['SPECTRAL_WINDOW_ID'] == spw_id:
                    pol_type.append(row['POLARIZATION_TYPE'])
            if spw_id == -1 or len(pol_type) != 0:
                break
        if len(pol_type) > 1:
            raise ValueError, 'Error getting Polarization type'
        elif len(pol_type) == 1:
            return [pol.upper() for pol in pol_type[0]]
        return '' # or None?

    def get_observation_time(self):
        """Return unix time of the observation. (Seconds since 1970.)"""
        main_table = table(self.msname, ack=False)
        time_0 = quantity(main_table[0]['TIME'], get_scalar_col_unit(main_table, 'TIME'))
        obs_time = time_0 - mod_jul_day
        set_unit(obs_time, 's')
        return obs_time.get_value()

    def _append_row(self, table_obj, row):
        nrow = len(table_obj)
        table_obj.addrows(1)
        table_obj[nrow] = row
        
    def add_flagcmd_entry(self, midpoint, interval, typ, reason, command):
        """Add flag entry to FLAG_CMD table."""
        flagcmd_table = self.flagCmd()
        # Input time units are in seconds, table units are probably
        # seconds too, but we'll handle it anyway.
        time = quantity(midpoint, 's') + mod_jul_day
        intvl = quantity(interval, 's')
        set_unit(time, get_scalar_col_unit(flagcmd_table, 'TIME'))
        set_unit(intvl, get_scalar_col_unit(flagcmd_table, 'INTERVAL'))
        row = {'TIME': time.get_value(),
            'INTERVAL': intvl.get_value(),
            'TYPE': typ, 'REASON': reason, 'COMMAND': command}
        self._append_row(flagcmd_table, row)

    def add_gain_entry(self, antenna_id, feed_id, spwindow_id, sorted_gains, curve):
        """Add gain curve entry to JIVE_GAINCURVE table."""
        gain_table = self.gainCurve()
        if curve['type'] == 'POLY':
            curve['coeffs'] = numpy.array(curve['coeffs'])
        else:
            curve['table'] = numpy.array(curve['table'])
        row = {'ANTENNA_ID': antenna_id, 'FEED_ID': feed_id,
            'SPECTRAL_WINDOW_ID': spwindow_id, 'GAIN': sorted_gains,
            'CURVE': curve}
        self._append_row(gain_table, row)

    def add_syscal_entry(self, antenna_id, feed_id, spwindow_id, secs, sorted_temperatures):
        """Add TSYS entry to SYSCAL table."""
        syscal_table = self.sysCal()
        time = quantity(secs, 's') + mod_jul_day
        set_unit(time, get_scalar_col_unit(syscal_table, 'TIME'))
        row = {'ANTENNA_ID': antenna_id, 'FEED_ID': feed_id,
            'SPECTRAL_WINDOW_ID': spwindow_id, 'TIME': time.get_value(),
            'TSYS': sorted_temperatures}
        self._append_row(syscal_table, row)
